# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.models.geojson
~~~~~~~~~~~~~~~~~~~~~~~~

This module contains the class implementation for the geojson

For more information about the API, please check out
https://www.mapillary.com/developer/api-documentation/.

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""

# Package
import json

# Local

# # Exceptions
from zensvi.download.mapillary.models.exceptions import InvalidOptionError


class Properties:
    """
    Representation for the properties in a GeoJSON

    :param properties: The properties as the input
    :type properties: dict

    :raises InvalidOptionError: Raised when the geojson passed is the invalid type - not a dict

    :return: A class representation of the model
    :rtype: mapillary.models.geojson.Properties
    """

    def __init__(self, *properties, **kwargs) -> None:
        """
        Initializing Properties constructor

        :param properties: Key value pair passed as list
        :type properties: list

        :param kwargs: The kwargs given to assign as properties
        :type kwargs: dict

        :return: The object created
        """

        # Validate that the geojson passed is indeed a dictionary
        if not isinstance(properties, dict):
            # Raise InvalidOptionError
            InvalidOptionError(
                # The parameter that caused the exception
                param="Properties.__init__.properties",
                # The invalid value passed
                value=properties,
                # The keys that should be passed instead
                options=["dict"],
            )

        for item in properties:
            for key in item:
                setattr(self, key, item[key])
        for key in kwargs:
            setattr(self, key, kwargs[key])

    def to_dict(self):
        """Return the dictionary representation of the Properties"""

        attr_representation = [
            key for key in dir(self) if not key.startswith("__") and key != "to_dict"
        ]

        return {key: getattr(self, key) for key in attr_representation}

    def __str__(self):
        """Return the informal string representation of the Properties"""

        attr_representation = [
            key for key in dir(self) if not key.startswith("__") and key != "to_dict"
        ]

        attr_key_value_pair = {key: getattr(self, key) for key in attr_representation}

        return f"{attr_key_value_pair}"

    def __repr__(self):
        """Return the formal string representation of the Properties"""

        attr_representation = [
            key for key in dir(self) if not key.startswith("__") and key != "to_dict"
        ]

        attr_key_value_pair = {key: getattr(self, key) for key in attr_representation}

        return f"{attr_key_value_pair}"


class Coordinates:
    """
    Representation for the coordinates in a geometry for a FeatureCollection

    :param longitude: The longitude of the coordinate set
    :type longitude: float

    :param latitude: The latitude of the coordinate set
    :type latitude: float

    :raises InvalidOptionError: Raised when invalid data types are passed to the coordinate set

    :return: A class representation of the Coordinates set
    :rtype: mapillary.models.geojson.Coordinates
    """

    def __init__(self, longitude: float, latitude: float) -> None:
        """
        Initializing Coordinates constructor

        :param longitude: The longitude of the coordinate set
        :type longitude: float

        :param latitude: The latitude of the coordinate set
        :type latitude: float
        """

        # Validate that the longitude passed is indeed a float
        if not isinstance(longitude, float):
            # Raise InvalidOptionError
            InvalidOptionError(
                # The parameter that caused the exception
                param="Coordinates.__init__.longitude",
                # The invalid value passed
                value=longitude,
                # The keys that should be passed instead
                options=["float"],
            )

        # Validate that the latitude passed is indeed a float
        if not isinstance(latitude, float):
            # Raise InvalidOptionError
            InvalidOptionError(
                # The parameter that caused the exception
                param="Coordinates.__init__.latitude",
                # The invalid value passed
                value=latitude,
                # The keys that should be passed instead
                options=["float"],
            )

        self.longitude = longitude
        self.latitude = latitude

    def to_list(self):
        """Return the list representation of the Coordinates"""

        return [self.longitude, self.latitude]

    def to_dict(self):
        """Return the dictionary representation of the Coordinates"""

        return {"lng": self.longitude, "lat": self.latitude}

    def __str__(self):
        """Return the informal string representation of the Coordinates"""

        return f"{self.longitude}, {self.latitude}"

    def __repr__(self) -> str:
        """Return the formal string representation of the Coordinates"""

        return f"{self.longitude}, {self.latitude}"


class Geometry:
    """
    Representation for the geometry in a GeoJSON

    :param geometry: The geometry as the input
    :type geometry: dict

    :raises InvalidOptionError: Raised when the geometry passed is the invalid type - not a dict

    :return: A class representation of the model
    :rtype: mapillary.models.geojson.Geometry
    """

    def __init__(self, geometry: dict) -> None:
        """
        Initializing Geometry constructor

        :param geometry: The geometry object for creation
        :type geometry: dict
        """

        # Validate that the geojson passed is indeed a dictionary
        if not isinstance(geometry, dict):
            # Raise InvalidOptionError
            InvalidOptionError(
                # The parameter that caused the exception
                param="Geometry.__init__.geometry",
                # The invalid value passed
                value=geometry,
                # The keys that should be passed instead
                options=["dict"],
            )

        # Setting the type of the selected geometry
        self.type: str = geometry["type"]

        # Setting the coordinates of the geometry
        self.coordinates: Coordinates = Coordinates(
            geometry["coordinates"][0], geometry["coordinates"][1]
        )

    def to_dict(self):
        """Return dictionary representation of the geometry"""

        return {"type": self.type, "coordinates": self.coordinates.to_list()}

    def __str__(self):
        """Return the informal string representation of the Geometry"""

        return f"{{'type': {self.type}, 'coordinates': {self.coordinates.to_list()}}}"

    def __repr__(self):
        """Return the formal string representation of the Geometry"""

        return f"{{'type': {self.type}, 'coordinates': {self.coordinates.to_list()}}}"


class Feature:
    """Representation for a feature in a feature list

    :param feature: The GeoJSON as the input
    :type feature: dict

    :raises InvalidOptionError: Raised when the geojson passed is the invalid type - not a dict

    :return: A class representation of the model
    :rtype: mapillary.models.geojson.Feature
    """

    def __init__(self, feature: dict) -> None:
        """
        Initializing Feature constructor

        :param feature: Feature JSON
        :type feature: dict
        """

        # Validate that the geojson passed is indeed a dictionary
        if not isinstance(feature, dict):
            # If not, raise `InvalidOptionError`
            InvalidOptionError(
                # The parameter that caused the exception
                param="Feature.__init__.feature",
                # The invalid value passed
                value=feature,
                # The type of value that should be passed instead
                options=["dict"],
            )

        # Setting the type of the selected FeatureList
        self.type = "Feature"

        # Setting the `geometry` property
        self.geometry = Geometry(feature["geometry"])

        # Setting the `properties` property
        self.properties = Properties(feature["properties"])

    def to_dict(self) -> dict:
        """Return the dictionary representation of the Feature"""

        return {
            "type": self.type,
            "geometry": self.geometry.to_dict(),
            "properties": self.properties.to_dict(),
        }

    def __str__(self) -> str:
        """Return the informal string representation of the Feature"""

        return (
            f"{{"
            f"'type': '{self.type}', "
            f"'geometry': {self.geometry}, "
            f"'properties': {self.properties}"
            f"}}"
        )

    def __repr__(self) -> str:
        """Return the formal string representation of the Feature"""

        return (
            f"{{"
            f"'type': {self.type}, "
            f"'geometry': {self.geometry}, "
            f"'properties': {self.properties}"
            f"}}"
        )


class GeoJSON:
    """Representation for a geojson

    :param geojson: The GeoJSON as the input
    :type geojson: dict

    :raises InvalidOptionError: Raised when the geojson passed is the invalid type - not a dict

    :return: A class representation of the model
    :rtype: mapillary.models.geojson.GeoJSON

    Usage::

        >>> import mapillary as mly
        >>> from models.geojson import GeoJSON
        >>> mly.interface.set_access_token('MLY|XXX')
        >>> data = mly.interface.get_image_close_to(longitude=31, latitude=31)
        >>> geojson = GeoJSON(geojson=data)
        >>> type(geojson)
        ... <class 'mapillary.models.geojson.GeoJSON'>
        >>> type(geojson.type)
        ... <class 'str'>
        >>> type(geojson.features)
        ... <class 'list'>
        >>> type(geojson.features[0])
        ... <class 'mapillary.models.geojson.Feature'>
        >>> type(geojson.features[0].type)
        ... <class 'str'>
        >>> type(geojson.features[0].geometry)
        ... <class 'mapillary.models.geojson.Geometry'>
        >>> type(geojson.features[0].geometry.type)
        ... <class 'str'>
        >>> type(geojson.features[0].geometry.coordinates)
        ... <class 'list'>
        >>> type(geojson.features[0].properties)
        ... <class 'mapillary.models.geojson.Properties'>
        >>> type(geojson.features[0].properties.captured_at)
        ... <class 'int'>
        >>> type(geojson.features[0].properties.is_pano)
        ... <class 'str'>
    """

    def __init__(self, geojson: dict) -> None:
        """Initializing GeoJSON constructor"""

        # Validate that the geojson passed is indeed a dictionary
        if isinstance(geojson, dict):

            # The GeoJSON should only contain the keys of `type`, `features`, if not empty,
            # raise exception
            if [key for key in geojson.keys() if key not in ["type", "features"]]:
                # Raise InvalidOptionError
                InvalidOptionError(
                    # The parameter that caused the exception
                    param="GeoJSON.__init__.geojson",
                    # The invalid value passed
                    value=geojson,
                    # The keys that should be passed instead
                    options=["type", "features"],
                )

        # If the GeoJSON is not of type dictionary
        else:

            # Raise InvalidOptionError
            InvalidOptionError(
                # The parameter that caused the exception
                param="GeoJSON.__init__.geojson",
                # The invalid value passed
                value=geojson,
                # The keys that should be passed instead
                options=["type", "features"],
            )

        # Validate that the geojson passed is indeed a dictionary
        if not isinstance(geojson["features"], list):
            # If not, raise InvalidOptionError
            InvalidOptionError(
                # The parameter that caused the exception
                param="FeatureList.__init__.geojson['features']",
                # The invalid value passed
                value=geojson["features"],
                # The type of the value that should be passed
                options=["list"],
            )

        # Setting the type parameter
        self.type: str = geojson["type"]

        # Setting the list of features
        self.features: list = (
            [Feature(feature=feature) for feature in geojson["features"]]
            if (geojson["features"] != []) or (geojson["features"] is not None)
            else []
        )

    def append_features(self, features: list) -> None:
        """
        Given a feature list, append it to the GeoJSON object

        :param features: A feature list
        :type features: list

        :return: None
        """

        # Iterating over features
        for feature in features:

            # Appending the feature to the GeoJSON
            self.append_feature(feature)

    def append_feature(self, feature_inputs: dict) -> None:
        """
        Given a feature dictionary, append it to the GeoJSON object

        :param feature_inputs: A feature as dict
        :type feature_inputs: dict

        :return: None
        """

        # Converting to a feature object
        feature = Feature(feature=feature_inputs)

        # If the feature does not already exist in self.features
        if feature not in self.features:

            # Append it
            self.features.append(feature)

    def encode(self) -> str:
        """
        Serializes the GeoJSON object

        :return: Serialized GeoJSON
        """

        return json.dumps(self.__dict__)

    def to_dict(self):
        """Return the dict format representation of the GeoJSON"""

        return {
            "type": self.type,
            "features": [feature.to_dict() for feature in self.features]
            if self.features != []
            else [],
        }

    def __str__(self):
        """Return the informal string representation of the GeoJSON"""

        return f"{{'type': '{self.type}', 'features': {self.features}}}"

    def __repr__(self):
        """Return the formal string representation of the GeoJSON"""

        return f"{{'type': '{self.type}', 'features': {self.features}}}"



# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.utils.extract
=======================

This module deals with extracting multiple fields nested within a GeoJSON packet.

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""


def extract_properties(geojson: dict, properties: list) -> dict:
    """
    Extracts specific properties from a complete GeoJSON

    :param geojson: GeoJSON object
    :type geojson: dict

    :param properties: A list of properties to extract
    :type properties: list

    :return: The extracted fields as a dict
    :rtype: dict

    Usage::

        >>> from utils.extract import extract_properties
        >>> extract_properties(geojson={"type":"FeatureCollection","features":[{"type":"Feature",
        ... "geometry":{"type":"Point","coordinates":[-80.12991070747375,25.787652114106322]},
        ... "properties":{"captured_at":1540386861135, "compass_angle":252.04260253906,"id":
        ... 1274987139570038,"is_pano":'False',"sequence_id":"Vf8Iwxx5SemxI7_b_7J5Kw"}},{"type":
        ... "Feature","geometry":{"type":"Point","coordinates":[-80.13223886489868,
        ... 25.78756517066695]}, "properties":{"captured_at":1422989164000,"compass_angle":
        ... 89.781,"id":169629268373019,"is_pano": "True","sequence_id":"dqjuprkOwUnmdEVt5gx-Iw"}}]}
        ... , properties=['id']) # id most likely exists
    """

    extracted_fields = {}

    for entry in properties:
        extracted_fields[entry] = []

    for feature in geojson["features"]:
        for entry in properties:
            if entry in feature["properties"]:
                extracted_fields[entry].append(feature["properties"][entry])

    return extracted_fields



# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.utils.__init__
========================

This package contains all the internal utilities used within the Mapillary python SDK.

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""

from . import auth  # noqa: F401
from . import extract  # noqa: F401
from . import filter  # noqa: F401
from . import format  # noqa: F401
from . import time  # noqa: F401
from . import verify  # noqa: F401



# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.models.api.vector_tiles
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module contains the adapter design for the Vector Tiles API of Mapillary API v4, Vector tiles
provide an easy way to visualize vast amounts of data. Vector tiles support filtering and querying
rendered features. Mapillary vector tiles follow the Mapbox tile specification.

For more information, please check out https://www.mapillary.com/developer/api-documentation/.

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""

# Package imports
from vt2geojson.tools import vt_bytes_to_geojson
import mercantile

# Local imports
# # Config
from zensvi.download.mapillary.config.api.vector_tiles import VectorTiles

# # Client import
from zensvi.download.mapillary.models.client import Client

# # Exception handling
from zensvi.download.mapillary.models.exceptions import InvalidOptionError

# # Models
from zensvi.download.mapillary.models.geojson import GeoJSON


class VectorTilesAdapter(object):
    """
    Adapter model for dealing with the VectorTiles API, through the DRY principle. The
    VectorTilesAdapter class can be instantiated in the controller modules, providing an
    abstraction layer that uses the Client class, endpoints provided by the API v4 under
    `/config/api/vector_tiles.py`.

    It performs parsing, handling of layers, properties, and fields to make it easier to
    write higher level logic for extracing information, and lets developers to focus only
    on writing the high level business logic without having to repeat the process of parsing
    and using libraries such as `mercantile`, 'vt2geojson' and others, caring only about
    inputs/outputs

    Usage::

        >>> import mapillary
        >>> from mapillary.models.api.vector_tiles import VectorTilesAdapter
        >>> latitude, longitude = 30, 31
        >>> VectorTilesAdapter().fetch_layer(layer="image", zoom=14, longitude=longitude,
        ...     latitude=latitude,
        ... )
        >>> VectorTilesAdapter().fetch_layer(layer="sequence", zoom=10, longitude=longitude,
        ...     latitude=latitude,
        ... )
        >>> VectorTilesAdapter().fetch_layer(layer="overview", zoom=3, longitude=longitude,
        ...     latitude=latitude,
        ... )
    """

    # FOR DEVELOPERS
    # Future changes here will most likely work on the following aspects,
    # 1. Adjusting zoom levels
    # 1.1. Most likely changes in __zoom_range_check
    # 2. The zoom levels themselves
    # 2.1. Most likely changes in __init__
    # 3. Preprocessing steps depending on the layer targeted
    # 3.1. Most likely changes in __preprocess_layer, __preprocess_computed_layer,
    # __preprocess_features

    def __init__(self) -> None:
        """Initializing VectorTilesAdapter constructor"""

        # Initialize client object
        self.client = Client()

        # * See "FOR DEVELOPERS (2, 2.1)"

        # Setting the max zoom value
        self.__min_zoom = 0

        # Setting the min zoom value
        self.__max_zoom = 14

    def fetch_layer(
        self, layer: str, longitude: float, latitude: float, zoom: int = 14
    ) -> dict:
        """
        Fetches an image tile layer depending on the coordinates, and the layer selected
        along with the zoom level

        :param layer: Either 'overview', 'sequence', 'image'
        :type layer: str

        :param longitude: The longitude of the coordinates
        :type longitude: float

        :param latitude: The latitude of the coordinates
        :type latitude: float

        :param zoom: The zoom level, [0, 14], inclusive
        :type zoom: int

        :return: A GeoJSON for that specific layer and the specified zoom level
        :rtype: dict
        """

        # Checking if the correct parameters are passed
        VectorTilesAdapter.__check_parameters(longitude=longitude, latitude=latitude)

        # Check for the correct zoom values against the layer specified
        self.__zoom_range_check(layer=layer, zoom=zoom)

        # Return the results of the layer after preprocessing steps
        return self.__preprocess_layer(
            # The layer to retrieve from
            layer=layer,
            # Turn coordinates into a tile
            tile=mercantile.tile(lng=longitude, lat=latitude, zoom=zoom),
            # The zoom level
            zoom=zoom,
        )

    def fetch_computed_layer(
        self, layer: str, zoom: int, longitude: float, latitude: float
    ):
        """
        Same as `fetch_layer`, but gets in return computed tiles only.
        Depends on the layer, zoom level, longitude and the latitude specifications

        :param layer: Either 'overview', 'sequence', 'image'
        :type layer: str

        :param zoom: The zoom level, [0, 14], inclusive
        :type zoom: int

        :param longitude: The longitude of the coordinates
        :type longitude: float

        :param latitude: The latitude of the coordinates
        :type latitude: float

        :return: A GeoJSON for that specific layer and the specified zoom level
        :rtype: dict
        """

        # Checking if the correct parameters are passed
        VectorTilesAdapter.__check_parameters(longitude=longitude, latitude=latitude)

        # Check for the correct zoom values against the layer specified
        self.__zoom_range_check(layer=layer, zoom=zoom)

        # Return the results of the layer after preprocessing steps
        return self.__preprocess_computed_layer(
            # The layer to retrieve from
            layer=layer,
            # Turn coordinates into a tile
            tile=mercantile.tile(lng=longitude, lat=latitude, zoom=zoom),
            # The zoom level
            zoom=zoom,
        )

    def fetch_features(
        self, feature_type: str, zoom: int, longitude: float, latitude: float
    ):
        """
        Fetches specified features from the coordinates with the appropriate zoom level

        :param feature_type: Either `point`, or `tiles`
        :type feature_type: str

        :param zoom: The zoom level
        :type zoom: int

        :param longitude: The longitude of the coordinates
        :type longitude: float

        :param latitude: The latitude of the coordinates
        :type latitude: float

        :return: A GeoJSON for that specific layer and the specified zoom level
        :rtype: dict
        """

        # Checking if the correct parameters are passed
        VectorTilesAdapter.__check_parameters(longitude=longitude, latitude=latitude)

        # Check for the correct zoom values against the layer specified
        self.__zoom_range_check(layer="map", zoom=zoom)

        # Return the results of the layer after preprocessing steps
        return self.__preprocess_features(
            # The feature type to retrieve for
            feature_type=feature_type,
            # Turn coordinates into a tile
            tile=mercantile.tile(lng=longitude, lat=latitude, zoom=zoom),
            # The zoom level
            zoom=zoom,
        )

    def fetch_layers(
        self,
        coordinates: "list[list]",
        layer: str = "image",
        zoom: int = 14,
        is_computed: bool = False,
    ) -> GeoJSON:
        """
        Fetches multiple vector tiles based on a list of multiple coordinates in a listed format

        :param coordinates: A list of lists of coordinates to get the vector tiles for
        :type coordinates: "list[list]"

        :param layer: Either "overview", "sequence", "image", "traffic_sign", or "map_feature",
            defaults to "image"
        :type layer: str

        :param zoom: the zoom level [0, 14], inclusive. Defaults to 14
        :type zoom: int

        :param is_computed: Will to be fetched layers be computed? Defaults to False
        :type is_computed: bool

        :return: A geojson with merged features from all unique vector tiles
        :rtype: dict
        """

        # Check for the correct zoom values against the layer specified
        self.__zoom_range_check(layer=layer, zoom=zoom)

        # The output resultant geojson
        geojson: GeoJSON = GeoJSON(
            geojson={"type": "FeatureCollection", "features": []}
        )

        # A list of tiles that are either confined within or intersect with the bbox
        tiles = list(
            mercantile.tiles(
                west=coordinates[0],
                south=coordinates[1],
                east=coordinates[2],
                north=coordinates[3],
                zooms=zoom,
            )
        )

        print(
            f'[Vector Tiles API] Fetching {len(tiles)} {"tiles" if len(tiles) > 1 else "tile"}'
            "for images ..."
        )

        for tile in tiles:

            if is_computed:
                result = (
                    self.__preprocess_computed_layer(
                        # The layer to retrieve from
                        layer=layer,
                        # Turn coordinates into a tile
                        tile=tile,
                        # The zoom level
                        zoom=zoom,
                    )
                )["features"]
            else:
                result = (
                    self.__preprocess_layer(
                        # The layer to retrieve from
                        layer=layer,
                        # Turn coordinates into a tile
                        tile=tile,
                        # The zoom level
                        zoom=zoom,
                    )
                )["features"]

            geojson.append_features(result)

        return geojson

    def fetch_map_features(
        self,
        coordinates: "list[list]",
        feature_type: str,
        zoom: int = 14,
    ) -> GeoJSON:
        """
        Fetches map features based on a list Polygon object

        :param coordinates: A list of lists of coordinates to get the map features for
        :type coordinates: "list[list]"

        :param feature_type: Either "point", "traffic_signs", defaults to "point"
        :type feature_type: str

        :param zoom: the zoom level [0, 14], inclusive. Defaults to 14
        :type zoom: int

        :return: A geojson with merged features from all unique vector tiles
        :rtype: dict
        """

        # Check for the correct zoom values against the layer specified
        self.__zoom_range_check(layer="map_feature", zoom=zoom)

        # The output resultant geojson
        geojson: GeoJSON = GeoJSON(
            geojson={"type": "FeatureCollection", "features": []}
        )

        # A list of tiles that are either confined within or intersect with the bbox
        tiles = list(
            mercantile.tiles(
                west=coordinates[0],
                south=coordinates[1],
                east=coordinates[2],
                north=coordinates[3],
                zooms=zoom,
            )
        )

        print(
            f'[Vector Tiles API] Fetching {len(tiles)} {"tiles" if len(tiles) > 1 else "tile"}'
            "for map features ..."
        )

        for tile in tiles:
            geojson.append_features(
                self.__preprocess_features(
                    # The layer to retrieve from
                    feature_type=feature_type,
                    # Turn coordinates into a tile
                    tile=tile,
                    # The zoom level
                    zoom=zoom,
                )["features"]
            )

        return geojson

    @staticmethod
    def __check_parameters(
        longitude: float,
        latitude: float,
    ):
        """
        Range checking for the parameters of longitude, latitude, layer, zoom

        :param longitude: The longitude of the coordinates
        :type longitude: float

        :param latitude: The latitude of the coordinates
        :type latitude: float

        :return: A GeoJSON for that specific layer and the specified zoom level
        :rtype: dict
        """

        # Lng, Lat ranges, https://docs.mapbox.com/help/glossary/lat-lon/

        # If lng not in the range [-180, 180], inclusive
        if longitude <= -180 or longitude >= 180:
            # Raise exception
            raise InvalidOptionError(
                param="longitude", value=longitude, options=[-180, 180]
            )

        # If lat not in the range [-90, 90], inclusive
        if latitude <= -90 or latitude >= 90:
            # Raise exception
            raise InvalidOptionError(
                param="latitude", value=latitude, options=[-180, 180]
            )

    def __zoom_range_check(self, layer: str, zoom: int):

        """
        Checks for the correct zoom values for te specified layer

        :param layer: Either 'overview', 'sequence', 'image', or 'map'
        :type layer: str

        :param zoom: The zoom levels,
        :type zoom: int

        :raises InvalidOptionError: Invalid option passed

        :return: A GeoJSON for the return object
        :rtype: dict
        """

        # If zoom is not in the valid range of values
        if zoom < self.__min_zoom or zoom > self.__max_zoom:
            # Raise an exception for the invalid values passed
            raise InvalidOptionError(
                # Parameters accordingly
                param="zoom",
                value=zoom,
                options=[_ for _ in range(self.__min_zoom, self.__max_zoom + 1)],
            )

        # If layer was specified to be 'overview'
        if layer == "overview":

            # If zoom is not in the given range of values
            if zoom not in [_ for _ in range(self.__min_zoom, 5 + 1)]:
                # Raise an exception for the invalid zoom value
                raise InvalidOptionError(
                    # Parameters accordingly
                    param="zoom",
                    value=zoom,
                    options=[_ for _ in range(self.__min_zoom, 5 + 1)],
                )

        # If layer was specified to be 'sequence'
        elif layer == "sequence":

            # IF zoom not in the given range of values
            if zoom not in [_ for _ in range(6, 14 + 1)]:
                # Raise an exception for the invalid zoom value
                raise InvalidOptionError(
                    # Parameters accordingly
                    param="zoom",
                    value=zoom,
                    options=[_ for _ in range(6, 14 + 1)],
                )

        # If layer was specified to be 'image' or 'map'
        elif layer == "image" or layer == "map_feature" or layer == "traffic_sign":

            # If zoom was not 14
            if zoom != 14:
                # Raise an exception for the invalid zoom value
                raise InvalidOptionError(param="zoom", value=zoom, options=[14])

        # Else, an invalid layer string was passed
        else:

            # Raise a InvalidOptionError
            raise InvalidOptionError(
                # Parameters accordingly
                param="layer",
                value=layer,
                options=["overview", "sequence", "image"],
            )

    def __preprocess_layer(self, layer: str, tile: mercantile.Tile, zoom: int):
        """
        Preprocessing un-computed layers

        :param layer: Either 'overview', 'sequence', 'image'
        :type layer: str

        :param tile: The specified tile
        :type tile: mercantile.Tile

        :param zoom: The zoom level
        :type zoom: int

        :return: A GeoJSON
        :rtype: dict
        """

        # * See "FOR DEVELOPERS (3, 3.1)"

        # Extract the url depending upon the layer
        url = ""

        # For overview
        if layer == "overview":
            url = VectorTiles.get_overview_layer(x=tile[0], y=tile[1], z=zoom)

        # For sequence
        elif layer == "sequence":
            url = VectorTiles.get_sequence_layer(x=tile[0], y=tile[1], z=zoom)

        # For image
        elif layer == "image":
            url = VectorTiles.get_image_layer(x=tile[0], y=tile[1], z=zoom)

        # No 'else' for InvalidOptionError, as checking done previously in __zoom_range_check

        # Convert bytes to GeoJSON
        return vt_bytes_to_geojson(
            # Parameters, appropriately
            b_content=self.client.get(url).content,
            x=tile.x,
            y=tile.y,
            z=tile.z,
            layer=layer,
        )

    def __preprocess_computed_layer(self, layer: str, tile: mercantile.Tile, zoom: int):
        """
        Preprocessing computed layers

        :param layer: Either 'overview', 'sequence', 'image'
        :type layer: str

        :param tile: The specified tile
        :type tile: mercantile.Tile

        :param zoom: The zoom level
        :type zoom: int

        :return: A GeoJSON
        :rtype: dict
        """

        # * See "FOR DEVELOPERS (3, 3.1)"

        # Extracting url from specified VectorTiles endpoint
        url = ""

        # For overview
        if layer == "overview":
            url = VectorTiles.get_computed_overview_layer(x=tile[0], y=tile[1], z=zoom)

        # For sequence
        elif layer == "sequence":
            url = VectorTiles.get_computed_sequence_layer(x=tile[0], y=tile[1], z=zoom)

        # For image
        elif layer == "image":
            url = VectorTiles.get_computed_image_layer(x=tile[0], y=tile[1], z=zoom)

        # No 'else' for InvalidOptionError, as checking done previously in __zoom_range_check

        # Convert bytes to geojson
        return vt_bytes_to_geojson(
            # Parameters appropriately
            b_content=self.client.get(url).content,
            x=tile.x,
            y=tile.y,
            z=tile.z,
            layer=layer,
        )

    def __preprocess_features(
        self, feature_type: str, tile: mercantile.Tile, zoom: int
    ) -> dict:
        """
        Preprocess features

        :param feature_type: Either 'point', 'traffic_signs'
        :type feature_type: str

        :param tile: The specified tile
        :type tile: mercantile.Tile

        :param zoom: The zoom level
        :type zoom: int

        :return: A GeoJSON
        :rtype: dict
        """

        # * See "FOR DEVELOPERS (3, 3.1)"

        # Extracting url from specified VectorTiles endpoint

        # For 'point'
        if feature_type == "point":

            url = VectorTiles.get_map_feature_point(x=tile[0], y=tile[1], z=zoom)

        # For 'traffic_signs'
        elif feature_type == "traffic_signs":
            url = VectorTiles.get_map_feature_traffic_sign(x=tile[0], y=tile[1], z=zoom)

        # If both are not present
        else:
            # Raise an exception for the invalid values passed
            raise InvalidOptionError(
                # Parameters accordingly
                param="feature_type",
                value=feature_type,
                options=["point", "traffic_sign"],
            )

        # Convert bytes to GeoJSON, and return
        return vt_bytes_to_geojson(
            # Parameters appropriately
            b_content=self.client.get(url).content,
            x=tile.x,
            y=tile.y,
            z=tile.z,
            layer=None,
        )



# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.utils.filter
======================

This module contains the filter utilies for high level filtering logic

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""

import logging

import haversine
from geojson import Point, Feature

# Local imports
from zensvi.download.mapillary.utils.time import date_to_unix_timestamp
from shapely.geometry import shape

# Package imports
from turfpy.measurement import bearing

logger = logging.getLogger("pipeline-logger")


def pipeline_component(func, data: list, exception_message: str, args: list) -> list:
    """
    A pipeline component which is respnonsible for sending functional arguments over
    to the selected target function - throwing a warning in case of an exception

    :param func: The filter to apply
    :type func: function

    :param data: The list of features to filter
    :type data: list

    :param exception_message: The exception message to print
    :type exception_message: str

    :param args: Arguments
    :type args: list

    :return: The filtered feature list
    :rtype: list

    Usage::

        >>> # internally used in mapillary.utils.pipeline
    """

    try:
        return func(data, *args)
    except TypeError as exception:
        logger.warning(f"{exception_message}, {exception}. Arguments passed, {args}")
        return []


def pipeline(data: dict, components: list) -> list:
    """
    A pipeline component that helps with making filtering easier. It provides
    access to different filtering mechanism by simplying letting users
    pass in what filter they want to apply, and the arguments for that filter

    :param data: The GeoJSON to be filtered
    :type data: dict

    :param components: The list of filters to apply
    :type components: list

    :return: The filtered feature list
    :rtype: list

    Usage::

        >>> # assume variables 'data', 'kwargs'
        >>> pipeline(
        ...     data=data,
        ...     components=[
        ...         {"filter": "image_type", "tile": kwargs["image_type"]}
        ...         if "image_type" in kwargs
        ...         else {},
        ...         {"filter": "organization_id", "organization_ids": kwargs["org_id"]}
        ...         if "org_id" in kwargs
        ...         else {},
        ...         {
        ...             "filter": "haversine_dist",
        ...             "radius": kwargs["radius"],
        ...             "coords": [longitude, latitude],
        ...         }
        ...         if "radius" in kwargs
        ...         else 1000
        ...     ]
        ... )
    """

    # Python treats dict objects as passed reference, thus
    # in order to not modify the previous state, we make a local copy
    __data = data.copy()["features"]

    # A mapping of different filters possible
    function_mappings = {
        "filter_values": filter_values,
        "max_captured_at": max_captured_at,
        "min_captured_at": min_captured_at,
        "haversine_dist": haversine_dist,
        "image_type": image_type,
        "organization_id": organization_id,
        "features_in_bounding_box": features_in_bounding_box,
        "existed_at": existed_at,
        "existed_before": existed_before,
        "sequence_id": sequence_id,
        "compass_angle": compass_angle,
        "hits_by_look_at": hits_by_look_at,
        "in_shape": in_shape,
        # Simply add the mapping of a new function,
        # nothing else will really need to changed
    }

    # Going through each of the components
    for component in components:

        # If component is simply empty, continue to next
        # iteration
        if component == {}:
            continue

        # Send to pipeline component, return data to `__data`
        __data = pipeline_component(
            # Map function respectively using the function_mappings dictionary
            func=function_mappings[f'{component["filter"]}'],
            # Send over the data
            data=__data,
            # Specify the message on the exception thrown
            exception_message=f'[pipeline - {component["filter"]}] Filter not applied, '
            "exception thrown",
            # Except the filter name, select the rest as args
            args=tuple(list(component.values())[1:]),
        )

    # Return the data
    return __data


def max_captured_at(data: list, max_timestamp: str) -> list:
    """
    Selects only the feature items that are less
    than the max_timestamp

    :param data: The feature list
    :type data: list

    :param max_timestamp: The UNIX timestamp as the max time
    :type max_timestamp: str

    :return: Filtered feature list
    :rtype: list

    Usage::

        >>> max_captured_at([{'type': 'Feature', 'geometry':
        ... {'type': 'Point', 'coordinates': [30.98594605922699, 30.003757307208872]}, 'properties':
        ... { ... }, ...}], '2020-05-23')
    """

    return [
        feature
        for feature in data
        if feature["properties"]["captured_at"] <= date_to_unix_timestamp(max_timestamp)
    ]


def min_captured_at(data: list, min_timestamp: str) -> list:
    """
    Selects only the feature items that are less
    than the min_timestamp

    :param data: The feature list
    :type data: list

    :param min_timestamp: The UNIX timestamp as the max time
    :type min_timestamp: str

    :return: Filtered feature list
    :rtype: list

    Usage::

        >>> max_captured_at([{'type': 'Feature', 'geometry':
        ... {'type': 'Point', 'coordinates': [30.98594605922699, 30.003757307208872]}, 'properties':
        ... { ... }, ...}], '2020-05-23')
    """

    return [
        feature
        for feature in data
        if feature["properties"]["captured_at"] >= date_to_unix_timestamp(min_timestamp)
    ]


def features_in_bounding_box(data: list, bbox: dict) -> list:
    """
    Filter for extracting features only in a bounding box

    :param data: the features list to be checked
    :type data: list

    :param bbox: Bounding box coordinates

        Example::
            >>> {
            ...     'west': 'BOUNDARY_FROM_WEST',
            ...     'south': 'BOUNDARY_FROM_SOUTH',
            ...     'east': 'BOUNDARY_FROM_EAST',
            ...     'north': 'BOUNDARY_FROM_NORTH'
            ... }

    :type bbox: dict

    :return: Features that only exist within the bounding box selected for the given features list
        provided in the BBox functon
    :rtype: list
    """

    # define an empty geojson as output
    output = []

    # For each feature in the filtered_data
    for feature in data:

        # If feature exists in bounding box
        if (
            feature["geometry"]["coordinates"][0] < bbox["east"]
            and feature["geometry"]["coordinates"][0] > bbox["west"]
        ) and (
            feature["geometry"]["coordinates"][1] > bbox["south"]
            and feature["geometry"]["coordinates"][1] < bbox["north"]
        ):
            # Append feature to the output
            output.append(feature)

    return output


def filter_values(data: list, values: list, property: str = "value") -> list:
    """
    Filter the features based on the existence of a specified value
    in one of the property.

    *TODO*: Need documentation that lists the 'values', specifically, it refers to 'value'
    *TODO*: under 'Detection', and 'Map feature', related to issue #65

    :param data: The data to be filtered
    :type data: dict

    :param values: A list of values to filter by
    :type values: list

    :param property: The specific parameter to look into
    :type property: str

    :return: A feature list
    :rtype: dict
    """

    return [
        feature for feature in data if feature["properties"].get(property) in values
    ]


def existed_at(data: list, existed_at: str) -> list:
    """
    Whether the first_seen_at properly existed after a specified time period

    :param data: The feature list
    :type data: list

    :param existed_at: The UNIX timestamp
    :type existed_at: str

    :return: The feature list
    :rtype: list
    """

    return [
        feature
        for feature in data
        if feature["properties"]["first_seen_at"] > date_to_unix_timestamp(existed_at)
    ]


def existed_before(data: list, existed_before: str) -> list:
    """
    Whether the first_seen_at properly existed before a specified time period

    :param data: The feature list
    :type data: list

    :param existed_before: The UNIX timestamp
    :type existed_before: str

    :return: A feature list
    :rtype: list
    """

    return [
        feature
        for feature in data
        if feature["properties"]["first_seen_at"]
        <= date_to_unix_timestamp(existed_before)
    ]


def haversine_dist(data: dict, radius: float, coords: list, unit: str = "m") -> list:
    """
    Returns features that are only in the radius specified using the Haversine distance, from
    the haversine package

    :param data: The data to be filtered
    :type data: dict

    :param radius: Radius for coordinates to fall into
    :type radius: float

    :param coords: The input coordinates (long, lat)
    :type coords: list

    :param unit: Either 'ft', 'km', 'm', 'mi', 'nmi', see here https://pypi.org/project/haversine/
    :type unit: str

    :return: A feature list
    :rtype: list
    """

    # Define an empty list
    output = []

    # Go through the features
    for feature in data:

        # Reverse the order of the coords to (lon, lat) before feeding into the haversine function
        reversed_coords = coords[::-1]
        reversed_feature_coords = feature["geometry"]["coordinates"][::-1]

        # If the calculated haversine distance is less than the radius ...
        if (
            haversine.haversine(reversed_coords, reversed_feature_coords, unit=unit)
            < radius
        ):
            # ... append to the output
            output.append(feature)

    # Return the output
    return output



def image_type(data: list, image_type: str) -> list:
    """
    The parameter might be 'all' (both is_pano == true and false), 'pano' (is_pano == true only),
    or 'flat' (is_pano == false only)

    :param data: The data to be filtered
    :type data: list

    :param image_type: Either 'pano' (True), 'flat' (False), or 'all' (None)
    :type image_type: str

    :return: A feature list
    :rtype: list
    """

    # Checking what kind of parameter is passed
    bool_for_pano_filtering = (
        # Return true if type == 'pano'
        True
        if image_type == "pano"
        # Else false if type == 'flat'
        else False
    )

    # Return the images based on image type
    return [
        feature
        for feature in data
        if feature["properties"]["is_pano"] == bool_for_pano_filtering
    ]


def organization_id(data: list, organization_ids: list) -> list:
    """
    Select only features that contain the specific organization_id

    :param data: The data to be filtered
    :type data: dict

    :param organization_ids: The oragnization id(s) to filter through
    :type organization_ids: list

    :return: A feature list
    :rtype: dict
    """

    return [
        # Feature only if
        feature
        # through the feature in the data
        for feature in data
        # if the found org_id is in the list of organization_ids
        if "organization_id" in feature["properties"]
        and feature["properties"]["organization_id"] in organization_ids
    ]


def sequence_id(data: list, ids: list) -> list:
    """
    Filter out images that do not have the sequence_id in the list of ids

    :param data: The data to be filtered
    :type data: list

    :param ids: The sequence id(s) to filter through
    :type ids: list

    :return: A feature list
    :rtype: list
    """

    return [feature for feature in data if feature["properties"]["sequence_id"] in ids]


def compass_angle(data: list, angles: tuple = (0.0, 360.0)) -> list:
    """
    Filter out images that do not lie within compass angle range

    :param data: The data to be filtered
    :type data: list

    :param angles: The compass angle range to filter through
    :type angle: tuple of floats

    :return: A feature list
    :rtype: list
    """

    if len(angles) != 2:
        raise ValueError("Angles must be a tuple of length 2")
    if angles[0] > angles[1]:
        raise ValueError("First angle must be less than second angle")
    if angles[0] < 0.0 or angles[1] > 360.0:
        raise ValueError("Angles must be between 0 and 360")

    return [
        feature
        for feature in data
        if angles[0] <= feature["properties"]["compass_angle"] <= angles[1]
    ]


def is_looking_at(image_feature: Feature, look_at_feature: Feature) -> bool:
    """
    Return whether the image_feature is looking at the look_at_feature

    :param image_feature: The feature set of the image
    :type image_feature: dict

    :param look_at_feature: The feature that is being looked at
    :type look_at_feature: dict

    :return: Whether the diff is greater than 310, or less than 50
    :rtype: bool
    """

    # Pano accessible via the `get_image_layer`
    # in config/api/vector_tiles.py
    if image_feature["properties"]["is_pano"]:
        return True

    # Compass angle accessible via the `get_image_layer`
    # in config/api/vector_tiles.py
    if image_feature["properties"]["compass_angle"] < 0:
        return False

    # Getting the difference between the two provided GeoJSONs and the compass angle
    diff: int = (
        abs(
            bearing(start=image_feature, end=look_at_feature)
            - image_feature["properties"]["compass_angle"]
        )
        % 360
    )

    # If diff > 310 OR diff < 50
    return 310 < diff or diff < 50


def by_look_at_feature(image: dict, look_at_feature: Feature) -> bool:
    """
    Filter through the given image features and return only features with the look_at_feature

    :param image: The feature dictionary
    :type image: dict

    :param look_at_feature: Feature description
    :type look_at_feature: A WGS84 GIS feature, TurfPy

    :return: Whether the given feature is looking at the `look_at_features`
    :rtype: bool
    """

    # Converting the coordinates in coords
    coords = [image["geometry"]["coordinates"][0], image["geometry"]["coordinates"][1]]

    # Getting the feature using `Feature`, `Point` from TurfPy
    image_feature = Feature(
        geometry=Point(coords, {"compass_angle": image["properties"]["compass_angle"]})
    )

    image_feature["properties"] = image["properties"]

    # Does the `image_feature` look at the `look_at_feature`?
    return is_looking_at(image_feature, look_at_feature)


def hits_by_look_at(data: list, at: dict) -> list:
    """
    Whether the given data have any feature that look at the `at` coordinates

    :param data: List of features with an Image entity
    :type data: list

    :param at: The lng and lat coordinates

        Example::

            >>> {
            ...     'lng': 'longitude',
            ...     'lat': 'latitude'
            ... }

    :type at: dict

    :return: Filtered results of features only looking at `at`
    :rtype: list
    """

    # Converting the `at` into a Feature object from TurfPy
    at_feature = Feature(geometry=Point((at["lng"], at["lat"])))

    return list(filter(lambda image: by_look_at_feature(image, at_feature), data))


def in_shape(data: list, boundary) -> list:
    """
    Whether the given feature list lies within the shape

    :param data: A feature list to be filtered
    :type data: list

    :param boundary: Shapely helper for determining existence of point within a boundary
    :type boundary:

    :return: A feature list
    :rtype: list
    """

    # Generating output format
    output = []

    # Iterating over features
    for feature in data:

        # Extracting point from geometry feature
        point = shape(feature["geometry"])

        # Checking if point falls within the boundary using shapely.geometry.point.point
        if boundary.contains(point):
            # If true, append to output features
            output.append(feature)

    # Return output
    return output



# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.controller.rules.verify
=================================

This module implements the verification of the filters or keys passed to each of the controllers
under `./controllers` that implement the business logic functionalities of the Mapillary
Python SDK.

For more information, please check out https://www.mapillary.com/developer/api-documentation/

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""

# Local Imports
from zensvi.download.mapillary.models.exceptions import (
    InvalidBBoxError,
    InvalidKwargError,
    InvalidOptionError,
)
from zensvi.download.mapillary.config.api.entities import Entities
from zensvi.download.mapillary.models.client import Client

# Package Imports
import requests
import re


def international_dateline_check(bbox):
    if bbox["west"] > 0 and bbox["east"] < 0:
        return True
    return False


def bbox_validity_check(bbox):
    # longitude check
    if bbox["west"] < 180 or bbox["east"] > 180:
        raise InvalidBBoxError(message="Input values exceed their permitted limits")
    # lattitude check
    elif bbox["north"] > 90 or bbox["south"] < 90:
        raise InvalidBBoxError(message="Input values exceed their permitted limits")
    # longitude validity check
    elif bbox["west"] > bbox["east"]:
        # extra check for international dateline
        # it could either be an error or cross an internaitonal dateline
        # hence if it is passing the dateline, return True
        if international_dateline_check(bbox):
            new_east = bbox["east"] + 360
            bbox["east"] = new_east
            return bbox
        raise InvalidBBoxError(message="Invalid values")
    # lattitude validitiy check
    elif bbox["north"] < bbox["south"]:
        raise InvalidBBoxError(message="Invalid values")
    elif bbox["north"] == bbox["south"] and bbox["west"] == bbox["east"]:
        # checking for equal values to avoid flat box
        raise InvalidBBoxError(message="Invalid values")

    return True


def kwarg_check(kwargs: dict, options: list, callback: str) -> bool:
    """
    Checks for keyword arguments amongst the kwarg argument to fall into the options list

    :param kwargs: A dictionary that contains the keyword key-value pair arguments
    :type kwargs: dict

    :param options: A list of possible arguments in kwargs
    :type options: list

    :param callback: The function that called 'kwarg_check' in the case of an exception
    :type callback: str

    :raises InvalidOptionError: Invalid option exception

    :return: A boolean, whether the kwargs are appropriate or not
    :rtype: bool
    """

    if kwargs is not None:
        for key in kwargs.keys():
            if key not in options:
                raise InvalidKwargError(
                    func=callback,
                    key=key,
                    value=kwargs[key],
                    options=options,
                )

    # If 'zoom' is in kwargs
    if ("zoom" in kwargs) and (kwargs["zoom"] < 14 or kwargs["zoom"] > 17):
        # Raising exception for invalid zoom value
        raise InvalidOptionError(
            param="zoom", value=kwargs["zoom"], options=[14, 15, 16, 17]
        )

    # if 'image_type' is in kwargs
    if ("image_type" in kwargs) and (
        kwargs["image_type"] not in ["pano", "flat", "all"]
    ):
        # Raising exception for invalid image_type value
        raise InvalidOptionError(
            param="image_type",
            value=kwargs["image_type"],
            options=["pano", "flat", "all"],
        )

    # If all tests pass, return True
    return True


def image_check(kwargs) -> bool:
    """
    For image entities, check if the arguments provided fall in the right category

    :param kwargs: A dictionary that contains the keyword key-value pair arguments
    :type kwargs: dict
    """

    # Kwarg argument check
    return kwarg_check(
        kwargs=kwargs,
        options=[
            "min_captured_at",
            "max_captured_at",
            "radius",
            "image_type",
            "organization_id",
            "fields",
        ],
        callback="image_check",
    )


def resolution_check(resolution: int) -> bool:
    """
    Checking for the proper thumbnail size of the argument

    :param resolution: The image size to fetch for
    :type resolution: int

    :raises InvalidOptionError: Invalid thumbnail size passed raises exception

    :return: A check if the size is correct
    :rtype: bool
    """

    if resolution not in [256, 1024, 2048]:
        # Raising exception for resolution value
        raise InvalidOptionError(
            param="resolution", value=str(resolution), options=[256, 1024, 2048]
        )

    return True


def image_bbox_check(kwargs: dict) -> dict:
    """
    Check if the right arguments have been provided for the image bounding box

    :param kwargs: The dictionary parameters
    :type kwargs: dict

    :return: A final dictionary with the kwargs
    :rtype: dict
    """

    if kwarg_check(
        kwargs=kwargs,
        options=[
            "max_captured_at",
            "min_captured_at",
            "image_type",
            "compass_angle",
            "organization_id",
            "sequence_id",
            "zoom",
        ],
        callback="image_bbox_check",
    ):
        return {
            "max_captured_at": kwargs.get("max_captured_at", None),
            "min_captured_at": kwargs.get("min_captured_at", None),
            "image_type": kwargs.get("image_type", None),
            "compass_angle": kwargs.get("compass_angle", None),
            "sequence_id": kwargs.get("sequence_id", None),
            "organization_id": kwargs.get("organization_id", None),
        }


def sequence_bbox_check(kwargs: dict) -> dict:
    """
    Checking of the sequence bounding box

    :param kwargs: The final dictionary with the correct keys
    :type kwargs: dict

    :return: A dictionary with all the options available specifically
    :rtype: dict
    """

    if kwarg_check(
        kwargs=kwargs,
        options=[
            "max_captured_at",
            "min_captured_at",
            "image_type",
            "organization_id",
            "zoom",
        ],
        callback="sequence_bbox_check",
    ):
        return {
            "max_captured_at": kwargs.get("max_captured_at", None),
            "min_captured_at": kwargs.get("min_captured_at", None),
            "image_type": kwargs.get("image_type", None),
            "organization_id": kwargs.get("organization_id", None),
        }


def points_traffic_signs_check(kwargs: dict) -> dict:
    """
    Checks for traffic sign arguments

    :param kwargs: The parameters to be passed for filtering
    :type kwargs: dict

    :return: A dictionary with all the options available specifically
    :rtype: dict
    """

    if kwarg_check(
        kwargs=kwargs,
        options=["existed_at", "existed_before"],
        callback="points_traffic_signs_check",
    ):
        return {
            "existed_at": kwargs.get("existed_at", None),
            "existed_before": kwargs.get("existed_before", None),
        }


def valid_id(identity: int, image=True) -> None:
    """
    Checks if a given id is valid as it is assumed. For example, is a given id expectedly an
    image_id or not? Is the id expectedly a map_feature_id or not?

    :param identity: The ID passed
    :type identity: int

    :param image: Is the passed id an image_id?
    :type image: bool

    :raises InvalidOptionError: Raised when invalid arguments are passed

    :return: None
    :rtype: None
    """

    # IF image == False, and error_check == True, this becomes True
    # IF image == True, and error_check == False, this becomes True
    if image ^ is_image_id(identity=identity, fields=[]):
        # The EntityAdapter() sends a request to the server, checking
        # if the id is indeed an image_id, TRUE is so, else FALSE

        # Raises an exception of InvalidOptionError
        raise InvalidOptionError(
            param="id",
            value=f"ID: {identity}, image: {image}",
            options=[
                "ID is image_id AND image is True",
                "ID is map_feature_id AND image is False",
            ],
        )


def is_image_id(identity: int, fields: list = None) -> bool:
    """
    Checks if the id is an image_id

    :param identity: The id to be checked
    :type identity: int

    :param fields: The fields to be checked
    :type fields: list

    :return: True if the id is an image_id, else False
    :rtype: bool
    """

    try:
        res = requests.get(
            Entities.get_image(
                image_id=str(identity),
                fields=fields if fields != [] else Entities.get_image_fields(),
            ),
            headers={"Authorization": f"OAuth {Client.get_token()}"},
        )
        return res.status_code == 200

    except requests.HTTPError:
        return False


def check_file_name_validity(file_name: str) -> bool:
    """
    Checks if the file name is valid

    Valid file names are,

    - Without extensions
    - Without special characters
    - A-Z, a-z, 0-9, _, -

    :param file_name: The file name to be checked
    :type file_name: str

    :return: True if the file name is valid, else False
    :rtype: bool
    """

    string_check = re.compile("[@.!#$%^&*()<>?/}{~:]")  # noqa: W605
    if (
        # File name characters are not all ASCII
        not all(ord(c) < 128 for c in file_name)
        # File name characters contain special characters or extensions
        or string_check.search(file_name)
    ):
        print(
            f"File name: {file_name} is not valid. Please use only letters, numbers, dashes,"
            f" and underscores. \nDefaulting to: mapillary_CURRENT_UNIX_TIMESTAMP_"
        )
        return False
    return True



# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.models.config
~~~~~~~~~~~~~~~~~~~~~~~~

This module contains the Config class which sets up some global variables fo the duration of
the session that the SDK is in use for.

For more information, please check out https://www.mapillary.com/developer/api-documentation/.

- Copyright: (c) 2021 Facebook
- License: MIT License
"""

from .logger import Logger

Logger.setup_logger(name="mapillary.models.config")


class Config:
    """
    Config setup for the SDK

    Different parts of the SDK react differently depending on what is set

    Usage::

        >>> from mapillary.models.config import Config

    :param use_strict: If set to True, the SDK will raise an exception if an invalid arguments
    are sent to the functions in config.api calls. If set to False, the SDK will just log a warning.
    :type use_strict: bool
    :default use_strict: True
    """

    # Strict mode will raise exceptions when,
    # 1. Invalid arguments are passed to a function, such as those
    # files in the config.api directory. This can mean invalid date
    # formats, invalid IDs, invalid data types, etc.

    use_strict = True

    def __init__(self, use_strict: bool = True) -> None:
        """
        Initialize the Config class
        """

        self.use_strict = use_strict



import cv2
import numpy as np
import math
from pathlib import Path
from concurrent.futures import ThreadPoolExecutor, as_completed
from tqdm import tqdm

def xyz2lonlat(xyz):
    atan2 = np.arctan2
    asin = np.arcsin

    norm = np.linalg.norm(xyz, axis=-1, keepdims=True)
    xyz_norm = xyz / norm
    x = xyz_norm[..., 0:1]
    y = xyz_norm[..., 1:2]
    z = xyz_norm[..., 2:]

    lon = atan2(x, z)
    lat = asin(y)
    lst = [lon, lat]

    out = np.concatenate(lst, axis=-1)
    return out


def lonlat2XY(lonlat, shape):
    X = (lonlat[..., 0:1] / (2 * np.pi) + 0.5) * (shape[1] - 1)
    Y = (lonlat[..., 1:] / (np.pi) + 0.5) * (shape[0] - 1)
    lst = [X, Y]
    out = np.concatenate(lst, axis=-1)

    return out


class ImageTransformer:
    def __init__(self, dir_input, dir_output):
        if isinstance(dir_input, str):
            dir_input = Path(dir_input)
        elif not isinstance(dir_input, Path):
            raise TypeError("dir_input must be a str or Path object.")
        if isinstance(dir_output, str):
            dir_output = Path(dir_output)
        elif not isinstance(dir_output, Path):
            raise TypeError("dir_output must be a str or Path object.")
        self._dir_input = dir_input
        self._dir_output = dir_output
    
    @property
    def dir_input(self):
        return self._dir_input

    @dir_input.setter
    def dir_input(self, value):
        if isinstance(value, str):
            value = Path(value)
        elif not isinstance(value, Path):
            raise TypeError("dir_input must be a str or Path object.")
        self._dir_input = value

    @property
    def dir_output(self):
        return self._dir_output

    @dir_output.setter
    def dir_output(self, value):
        if isinstance(value, str):
            value = Path(value)
        elif not isinstance(value, Path):
            raise TypeError("dir_output must be a str or Path object.")
        self._dir_output = value

    def get_perspective(self, img, FOV, THETA, PHI, height, width):
        #
        # THETA is left/right angle, PHI is up/down angle, both in degree
        #

        f = 0.5 * width * 1 / np.tan(0.5 * FOV / 180.0 * np.pi)
        cx = (width - 1) / 2.0
        cy = (height - 1) / 2.0
        K = np.array([
            [f, 0, cx],
            [0, f, cy],
            [0, 0, 1],
        ], np.float32)
        K_inv = np.linalg.inv(K)

        x = np.arange(width)
        y = np.arange(height)
        x, y = np.meshgrid(x, y)
        z = np.ones_like(x)
        xyz = np.concatenate([x[..., None], y[..., None], z[..., None]], axis=-1)
        xyz = xyz @ K_inv.T

        y_axis = np.array([0.0, 1.0, 0.0], np.float32)
        x_axis = np.array([1.0, 0.0, 0.0], np.float32)
        R1, _ = cv2.Rodrigues(y_axis * np.radians(THETA))
        R2, _ = cv2.Rodrigues(np.dot(R1, x_axis) * np.radians(PHI))
        R = R2 @ R1
        xyz = xyz @ R.T
        lonlat = xyz2lonlat(xyz)
        XY = lonlat2XY(lonlat, shape=img.shape).astype(np.float32)
        persp = cv2.remap(img, XY[..., 0], XY[..., 1], cv2.INTER_LINEAR, borderMode=cv2.BORDER_WRAP)
        return persp
    
    def equidistant_fisheye(self, img):
        rows, cols, c = img.shape
        R = cols / (2 * math.pi)
        D = int(2 * R)
        cx, cy = R, R

        x, y = np.meshgrid(np.arange(D), np.arange(D))
        r = np.sqrt((x - cx) ** 2 + (y - cy) ** 2)
        theta = np.arctan2(y - cy, x - cx)

        xp = np.floor((theta + np.pi) * cols / (2 * np.pi)).astype(int)
        yp = np.floor(r * rows / R).astype(int)
        
        mask = r < R

        new_img = np.zeros((D, D, c), dtype=np.uint8)
        new_img.fill(255)
        new_img[y[mask], x[mask]] = img[yp[mask], xp[mask]]

        return new_img
    
    def orthographic_fisheye(self, img):
        rows, cols, c = img.shape
        R = cols / (2 * math.pi)
        D = int(2 * R)
        cx, cy = R, R

        x, y = np.meshgrid(np.arange(D), np.arange(D))
        r = np.sqrt((x - cx) ** 2 + (y - cy) ** 2)
        theta = np.arctan2(y - cy, x - cx)

        xp = np.floor((theta + np.pi) * cols / (2 * np.pi)).astype(int)
        yp = np.floor((2 / np.pi) * np.arcsin(r / R) * rows).astype(int)
        
        mask = r < R

        new_img = np.zeros((D, D, c), dtype=np.uint8)
        new_img.fill(255)
        new_img[y[mask], x[mask]] = img[yp[mask], xp[mask]]

        return new_img

    def stereographic_fisheye(self, img):
        rows, cols, c = img.shape
        R = cols / (2 * math.pi)
        D = int(2 * R)
        cx, cy = R, R

        x, y = np.meshgrid(np.arange(D), np.arange(D))
        r = np.sqrt((x - cx) ** 2 + (y - cy) ** 2)
        theta = np.arctan2(y - cy, x - cx)

        xp = np.floor((theta + np.pi) * cols / (2 * np.pi)).astype(int)
        yp = np.floor(2 * np.tan(r / (2 * R)) * rows).astype(int)
        
        # Clip the values of yp and xp to be within the valid range
        yp = np.clip(yp, 0, rows-1)
        xp = np.clip(xp, 0, cols-1)
        
        mask = r < R

        new_img = np.zeros((D, D, c), dtype=np.uint8)
        new_img.fill(255)
        new_img[y[mask], x[mask]] = img[yp[mask], xp[mask]]

        return new_img

    def equisolid_fisheye(self, img):
        rows, cols, c = img.shape
        R = cols / (2 * math.pi)
        D = int(2 * R)
        cx, cy = R, R

        x, y = np.meshgrid(np.arange(D), np.arange(D))
        r = np.sqrt((x - cx) ** 2 + (y - cy) ** 2)
        theta = np.arctan2(y - cy, x - cx)

        xp = np.floor((theta + np.pi) * cols / (2 * np.pi)).astype(int)
        yp = np.floor(2 * np.sin(r / (2 * R)) * rows).astype(int)
        
        mask = r < R

        new_img = np.zeros((D, D, c), dtype=np.uint8)
        new_img.fill(255)
        new_img[y[mask], x[mask]] = img[yp[mask], xp[mask]]

        return new_img

    def get_fisheye(self, img):
        rows, cols, c = img.shape
        R = int(cols / (2 * math.pi))
        D = R * 2
        cx = R
        cy = R

        # Create coordinate grids
        x, y = np.meshgrid(np.arange(D), np.arange(D))
        
        # Compute r and theta in a vectorized way
        r = np.sqrt((x - cx) ** 2 + (y - cy) ** 2)
        theta = np.arctan2(y - cy, x - cx) % (2 * math.pi)

        # Compute corresponding coordinates in the equirectangular image
        xp = np.floor(theta / (2 * math.pi) * cols).astype(int)
        yp = np.floor(r / R * rows).astype(int) - 1

        # Create a mask for pixels within the circle
        mask = r <= R

        # Apply mask to coordinate grids
        xp = xp[mask]
        yp = yp[mask]

        # Create new image and fill with white
        new_img = np.zeros((D, D, c), dtype=np.uint8)
        new_img.fill(255)

        # Copy pixels from original image to new image
        new_img[y[mask], x[mask]] = img[yp, xp]

        return new_img

    def transform_images(self, style_list=["perspective", "equidistant_fisheye", "orthographic_fisheye", "stereographic_fisheye", "equisolid_fisheye"], 
                    FOV = 90, theta = 90, aspects = (9, 16), show_size=100):
        # check if there's anything other than "perspective" and "fisheye"
        if not all(style in ["perspective", "equidistant_fisheye", "orthographic_fisheye", "stereographic_fisheye", "equisolid_fisheye"] for style in style_list):
            raise ValueError("Please input the correct image style. The correct image style should be 'perspective', 'equidistant_fisheye', 'orthographic_fisheye', 'stereographic_fisheye', or 'equisolid_fisheye'")

        # set image file extensions
        image_extensions = ['.jpg', '.jpeg', '.png', '.bmp', '.gif', '.tiff', '.tif', '.webp', '.ico', '.jfif', '.heic', '.heif']

        def run(path_input, path_output, show_size, style, theta, aspects, FOV):
            img_raw = cv2.imread(str(path_input), cv2.IMREAD_COLOR)
            if style == "equidistant_fisheye":
                if not path_output.exists():
                    img_new = self.equidistant_fisheye(img_raw)
                    cv2.imwrite(str(path_output), img_new)
            
            elif style == "orthographic_fisheye":
                if not path_output.exists():
                    img_new = self.orthographic_fisheye(img_raw)
                    cv2.imwrite(str(path_output), img_new)
                
            elif style == "stereographic_fisheye":
                if not path_output.exists():
                    img_new = self.stereographic_fisheye(img_raw)
                    cv2.imwrite(str(path_output), img_new)
            
            elif style == "equisolid_fisheye":
                if not path_output.exists():
                    img_new = self.equisolid_fisheye(img_raw)
                    cv2.imwrite(str(path_output), img_new)

            elif style == "perspective":
                num_images = 360 // theta  # Calculate the number of images based on theta
                thetas = [theta * i for i in range(num_images)]  # Calculate thetas based on step size

                for theta in thetas:
                    height = int(aspects[0] * show_size)
                    width = int(aspects[1] * show_size)
                    aspect_name = '%s--%s' % (aspects[0], aspects[1])
                    path_output_raw = path_output.with_name(f'{path_output.stem}_Direction_{theta}_FOV_{FOV}_aspect_{aspect_name}_raw.png')
                    if not path_output_raw.exists(): 
                        img_new = self.get_perspective(img_raw, FOV, theta, 0, height, width)
                        cv2.imwrite(str(path_output_raw), img_new)

        def process_image(dir_input, dir_output, name, show_size, style, theta, aspects, FOV):
            path_input = dir_input / name.name
            path_output = dir_output / (name.stem + ".png")
            return path_input, path_output, show_size, style, theta, aspects, FOV

        for current_style in style_list:
            dir_output = Path(self.dir_output) / current_style
            dir_output.mkdir(parents=True, exist_ok=True)

            with ThreadPoolExecutor() as executor:
                futures = [executor.submit(run, *process_image(self.dir_input, dir_output, name, show_size, current_style, theta, aspects, FOV)) \
                    for name in self.dir_input.rglob('*') if name.suffix.lower() in image_extensions]
                for future in tqdm(as_completed(futures), total=len(futures), desc=f"Converting to {current_style}"):
                    future.result()


# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.models.client
~~~~~~~~~~~~~~~~~~~~~~~

This module contains aims to serve as a generalization for all API requests within the Mapillary
Python SDK.

Over Authentication
!!!!!!!!!!!!!!!!!!!

1. All requests against https://graph.mapillary.com must be authorized. They require a client or
    user access tokens. Tokens can be sent in two ways,

    1. Using ?access_token=XXX query parameters. This is a preferred method for interacting with
        vector tiles. Using this method is STRONGLY discouraged for sending user access tokens
    2. Using a header such as Authorization: OAuth XXX, where XXX is the token obtained either
        through the OAuth flow that your application implements or a client token from
        https://mapillary.com/dashboard/developers.

For more information, please check out https://www.mapillary.com/developer/api-documentation/.

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""

import json
import logging
import os
import sys
from math import floor

import requests

# Exception imports
from zensvi.download.mapillary.models.exceptions import InvalidTokenError

# Basic logger setup
logger = logging.getLogger("mapillary.utils.client")

# stdout logger setup
stream_handler = logging.StreamHandler(sys.stdout)
logger.addHandler(stream_handler)

# Setting log_level to INFO
log_level = os.environ.get("LOG_LEVEL", "INFO").upper()

# Check if in DEBUG mode to show debugging output
if os.environ.get("DEBUG") == "1":
    log_level = "DEBUG"
try:
    logger.setLevel(log_level)
except ValueError:
    logger.setLevel(logging.INFO)
    logger.warning("LOG_LEVEL: invalid variable - Defaulting to: INFO")


class Client:
    """
    Client setup for API communication.

    All requests for the Mapillary API v4 should go through this class

    Usage::

        >>> client = Client(access_token='MLY|XXX')
        >>> # for entities endpoints
        >>> client.get(endpoint='endpoint specific path', entity=True, params={
        ...     'fields': ['id', 'value']
        ... })
        >>> # for tiles endpoint
        >>> client.get(endpoint='endpoint specific path', entity=False)
    """

    # User Access token will be set once and used throughout all requests
    # within the same session
    __access_token = ""

    def __init__(self) -> None:

        # Session object setup to be referenced across future API calls.
        self.session = requests.Session()

    @staticmethod
    def __check_token_validity(token):
        res = requests.get(
            "https://graph.mapillary.com/1933525276802129?fields=id",
            headers={"Authorization": f"OAuth {token}"},
        )

        if res.status_code == 401:
            res_content = json.loads(res.content)
            raise InvalidTokenError(
                res_content["error"]["message"],
                res_content["error"]["type"],
                res_content["error"]["code"],
                res_content["error"]["fbtrace_id"],
            )

    @staticmethod
    def get_token() -> str:
        """
        Gets the access token

        :return: The access token
        """

        return Client.__access_token

    @staticmethod
    def set_token(access_token: str) -> None:
        """
        Sets the access token

        :param access_token: The access token to be set
        """

        Client.__check_token_validity(access_token)

        Client.__access_token = access_token

    def _initiate_request(self, url: str, method: str, params: dict = None):
        """
        Private method - For internal use only.
        This method is responsible for making tailored API requests to the mapillary API v4.
        It generalizes the requests and ties them to the same session.

        :param url: The request endpoint - required
        :type url: str

        :param method: HTTP method to be used - required
        :type method: str

        :param params: Query parameters to be attached to the request - optional
        :type params: dict
        """

        request = requests.Request(method, url, params=params)

        # create a prepared request with the request and the session info merged
        prepped_req = self.session.prepare_request(request)

        # Log the prepped request before sending it.
        Client._pprint_request(prepped_req)

        # Sending the request
        res = self.session.send(prepped_req)

        # Log the responses
        Client._pprint_response(res)

        # Handling the response status codes
        if res.status_code == requests.codes.ok:

            try:
                logger.debug(f"Response: {res.json()}")

            except ValueError:
                return res

        elif res.status_code >= 400:

            logger.error(f"Server responded with a {str(res.status_code)} error!")

            try:
                logger.debug(f"Error details: {str(res.json())}")

            except ValueError:
                logger.debug(
                    "[Client - _initiate_request, ValueError] res.json() not available,"
                    "empty response"
                )

            res.raise_for_status()

        return res

    def get(self, url: str = None, params: dict = {}):
        """
        Make GET requests to both mapillary main endpoints

        :param url: The specific path of the request URL
        :type url: str

        :param params: Query parameters to be attached to the URL (Dict)
        :type params: dict
        """
        # Check if an endpoint is specified.
        if url is None:
            logger.error("You need to specify an endpoint!")
            return

        # Determine Authentication method based on the requested endpoint
        if "https://graph.mapillary.com" in url:
            self.session.headers.update(
                {"Authorization": f"OAuth {self.__access_token}"}
            )
        else:
            params["access_token"] = params.get("access_token", self.__access_token)

        return self._initiate_request(url=url, method="GET", params=params)

    @staticmethod
    def _pprint_request(prepped_req):
        """
        Format::

            Method endpoint: HTTP/version
            Host: host
            Header_key: Header_value
            Body

        :param prepped_req: The prepped request object

        Reference::

            1. 'https://github.com/michaeldbianchi/Python-API-Client-Boilerplate/blob/fd1c82be9e98e'
                '24730c4631ffc30068272386669/exampleClient.py#L202'
        """

        method = prepped_req.method
        url = prepped_req.url

        headers = "\n".join(f"{k}: {v}" for k, v in prepped_req.headers.items())
        # Print body if present or empty string if not
        body = prepped_req.body or ""

        logger.info(f"Requesting {method} to {url}")

        logger.debug(
            "{}\n{} {} HTTP/1.1\n{}\n\n{}".format(
                "-----------REQUEST-----------", method, url, headers, body
            )
        )

    @staticmethod
    def _pprint_response(res):
        """
        Format::

            HTTP/version status_code status_text
            Header_key: Header_value
            Body

        :param res: Response object returned from the API request

        Reference::

            1. 'https://github.com/michaeldbianchi/Python-API-Client-Boilerplate/blob/fd1c82be9e98e'
                '24730c4631ffc30068272386669/exampleClient.py#L230'
        """

        http_v0, http_v1 = list(str(res.raw.version))
        http_v = f"HTTP/{http_v0}.{http_v1}"
        status_code = res.status_code
        status_text = res.reason
        headers = "\n".join(f"{k}: {v}" for k, v in res.headers.items())
        body = res.text or ""

        # Convert timedelta to milliseconds
        elapsed = floor(res.elapsed.total_seconds() * 1000)

        logger.info(f"Response {status_code} {status_text} received in {elapsed}ms")

        logger.debug(
            "{}\n{} {} {}\n{}\n\n{}".format(
                "-----------RESPONSE-----------",
                http_v,
                status_code,
                status_text,
                headers,
                body,
            )
        )



import os
import random
import time
import datetime
import pandas as pd
from concurrent.futures import ThreadPoolExecutor, as_completed, ProcessPoolExecutor
import warnings
import requests
import cv2
import pkg_resources
from pathlib import Path
import geopandas as gpd
from tqdm import tqdm
from shapely.geometry import Point
import warnings
from shapely.errors import ShapelyDeprecationWarning
warnings.filterwarnings("ignore", category=ShapelyDeprecationWarning)
import csv
import glob
import shutil
import numpy as np
import osmnx as ox
from abc import ABC, abstractmethod
from shapely import wkt
from PIL import Image
from typing import List, Tuple, Union

import zensvi.download.mapillary.interface as mly
from zensvi.download.utils.imtool import ImageTool
from zensvi.download.utils.get_pids import panoids
from zensvi.download.utils.geoprocess import GeoProcessor
from zensvi.download.utils.helpers import standardize_column_names, create_buffer_gdf

# set logging level to warning
import logging
logging.getLogger('mapillary.utils.client').setLevel(logging.WARNING)

class BaseDownloader(ABC):
    @abstractmethod
    def __init__(self, log_path = None, distance = 1, grid = False, grid_size = 1):
        self._log_path = log_path
        self._distance = distance
        self._grid = grid
        self._grid_size = grid_size
        self._user_agents = self._get_ua()
        self._proxies = self._get_proxies()

    @property
    def log_path(self):
        return self._log_path    
    @log_path.setter
    def log_path(self,log_path):
        self._log_path = log_path
    
    @property
    def distance(self):
        return self._distance    
    @distance.setter
    def distance(self,distance):
        self._distance = distance
    
    @property
    def grid(self):
        return self._grid    
    @grid.setter
    def grid(self,grid):
        self._grid = grid
        
    @property
    def grid_size(self):
        return self._grid_size    
    @grid_size.setter
    def grid_size(self,grid_size):
        self._grid_size = grid_size

    @property
    def proxies(self):
        return self._proxies
    
    def _get_proxies(self):
        proxies_file = pkg_resources.resource_filename('zensvi.download.utils', 'proxies.csv')
        proxies = []
        # open with "utf-8" encoding to avoid UnicodeDecodeError
        with open(proxies_file, 'r', encoding = "utf-8") as f:
            reader = csv.DictReader(f)
            for row in reader:
                ip = row['ip']
                port = row['port']
                protocols = row['protocols']
                proxy_dict = {protocols: f"{ip}:{port}"}
                proxies.append(proxy_dict)
        return proxies
    
    @property
    def user_agents(self):
        return self._user_agents
    
    def _get_ua(self):
        user_agent_file = pkg_resources.resource_filename('zensvi.download.utils', 'UserAgent.csv')
        UA = []
        with open(user_agent_file, 'r') as f:
            for line in f:
                ua = {"user_agent": line.strip()}
                UA.append(ua)
        return UA
    
    def _log_write(self, pids):
        if self.log_path == None:
            return
        with open(self.log_path, 'a+') as fw:
            for pid in pids:
                fw.write(pid+'\n')
                
    def _check_already(self, all_panoids):
        # Get the set of already downloaded images
        name_r = set()
        for dirpath, dirnames, filenames in os.walk(self.panorama_output):
            for name in filenames:
                name_r.add(name.split(".")[0])

        # Filter the list of all panoids to only include those not already downloaded
        all_panoids = list(set(all_panoids) - name_r)
        return all_panoids
    
    def _read_pids(self, path_pid, start_date, end_date):
        pid_df = pd.read_csv(path_pid)
        # filter pids by date
        pid_df = self._filter_pids_date(pid_df, start_date, end_date)
        # get unique pids as a list
        pids = pid_df.iloc[:,0].unique().tolist()
        return pids

    @abstractmethod
    def _filter_pids_date(self, pid_df, start_date, end_date):
        pass

    @abstractmethod
    def download_svi(self, 
                    dir_output, 
                    lat=None, 
                    lon=None, 
                    input_csv_file="", 
                    input_shp_file = "", 
                    input_place_name = "", 
                    id_columns=None, 
                    buffer = 0,
                    update_pids = False,
                    start_date = None,
                    end_date = None):
        pass


class GSVDownloader(BaseDownloader):
    def __init__(self, gsv_api_key: str = None, log_path: str = None, distance: int = 1, grid: bool = False, grid_size: int = 1):
        """
        Google Street View Downloader class.

        Args:
            gsv_api_key (str, optional): Google Street View API key. Defaults to None.
            log_path (str, optional): Path to the log file. Defaults to None.
            distance (int, optional): Distance parameter for the GeoProcessor. Defaults to 1.
            grid (bool, optional): Grid parameter for the GeoProcessor. Defaults to False.
            grid_size (int, optional): Grid size parameter for the GeoProcessor. Defaults to 1.

        Raises:
            Warning: If gsv_api_key is not provided.
        """
        super().__init__(log_path, distance, grid, grid_size)
        if gsv_api_key == None:
            warnings.warn("Please provide your Google Street View API key to augment metadata.")
        self._gsv_api_key = gsv_api_key

    @property
    def gsv_api_key(self) -> str:
        """
        Property to get the Google Street View API key.

        Returns:
            str: Google Street View API key.
        """
        return self._gsv_api_key    
    
    @gsv_api_key.setter
    def gsv_api_key(self, gsv_api_key: str):
        """
        Setter to set the Google Street View API key.

        Args:
            gsv_api_key (str): Google Street View API key.
        """
        self._gsv_api_key = gsv_api_key

    def _augment_metadata(self, df):
        if self.cache_pids_augmented.exists():
            df = pd.read_csv(self.cache_pids_augmented)
            print("The augmented panorama IDs have been read from the cache")
            return df
        
        # Create a new directory called "augmented_metadata_checkpoints"
        dir_cache_augmented_metadata = self.dir_cache / 'augmented_pids'
        dir_cache_augmented_metadata.mkdir(parents=True, exist_ok=True)

        # Load all the checkpoint csv files
        checkpoints = glob.glob(str(dir_cache_augmented_metadata / '*.csv'))
        checkpoint_start_index = len(checkpoints)

        if checkpoint_start_index > 0:
            completed_rows = pd.concat([pd.read_csv(checkpoint) for checkpoint in checkpoints], ignore_index=True)
            completed_indices = completed_rows.index.unique()

            # Filter df to get remaining indices to augment metadata for
            df = df.loc[~df.index.isin(completed_indices)]

        def get_year_month(pid, proxies):
            url = "https://maps.googleapis.com/maps/api/streetview/metadata?pano={}&key={}".format(pid, self.gsv_api_key)
            while True:
                proxy = random.choice(proxies)
                try:
                    response = requests.get(url, proxies=proxy, timeout=5)
                    break
                except Exception as e:
                    print(f"Proxy {proxy} is not working. Exception: {e}")
                    continue

            response = response.json()
            if response['status'] == 'OK':
                # get year and month from date
                try:
                    date = response['date']
                    year = date.split("-")[0]
                    month = date.split("-")[1]
                except Exception:
                    year = None
                    month = None
                return {"year": year, "month": month}
            return {"year": None, "month": None}    

        def worker(row, proxies):
            panoid = row.panoid
            year_month = get_year_month(panoid, proxies)
            return row.Index, year_month

        batch_size = 1000  # Modify this to a suitable value
        num_batches = (len(df) + batch_size - 1) // batch_size

        for i in tqdm(range(num_batches), desc=f"Augmenting metadata by batch size {min(batch_size, len(df))}"):
            batch_df = df.iloc[i*batch_size : (i+1)*batch_size].copy()  # Copy the batch data to a new dataframe
            with ThreadPoolExecutor() as executor:
                batch_futures = {executor.submit(worker, row, self.proxies): row.Index for row in batch_df.itertuples()}
                for future in tqdm(as_completed(batch_futures), total=len(batch_futures), desc=f"Augmenting metadata for batch #{i+1}"):
                    row_index, year_month = future.result()
                    batch_df.at[row_index, 'year'] = year_month['year']
                    batch_df.at[row_index, 'month'] = year_month['month']

            # Save checkpoint for each batch
            batch_df.to_csv(f'{dir_cache_augmented_metadata}/checkpoint_batch_{checkpoint_start_index+i+1}.csv', index=False)
        
        # Merge all checkpoints into a single dataframe
        df = pd.concat([pd.read_csv(checkpoint) for checkpoint in glob.glob(str(dir_cache_augmented_metadata / '*.csv'))], ignore_index=True)

        # save the augmented metadata
        df.to_csv(self.cache_pids_augmented, index=False)
        # delete cache_lat_lon
        if self.cache_lat_lon.exists():
            self.cache_lat_lon.unlink() 
        # delete cache_pids_raw
        if self.cache_pids_raw.exists():
            self.cache_pids_raw.unlink()
        # delete the cache directory
        if dir_cache_augmented_metadata.exists():
            shutil.rmtree(dir_cache_augmented_metadata)
        return df

    def _get_pids_from_df(self, df, id_columns=None):
        # 1. Create a new directory called "pids" to store each batch pids
        dir_cache_pids = self.dir_cache / 'raw_pids'
        dir_cache_pids.mkdir(parents=True, exist_ok=True)

        # 2. Load all the checkpoint csv files
        checkpoints = glob.glob(str(dir_cache_pids / '*.csv'))
        checkpoint_start_index = len(checkpoints)
        if checkpoint_start_index > 0:
            dataframes = []
            for checkpoint in checkpoints:
                try:
                    df_checkpoint = pd.read_csv(checkpoint)
                    dataframes.append(df_checkpoint)
                except pd.errors.EmptyDataError:
                    print(f"Warning: {checkpoint} is empty and has been skipped.")
                    continue
            completed_rows = pd.concat(dataframes, ignore_index=True)

            completed_ids = completed_rows['lat_lon_id'].drop_duplicates()

            # Merge on the ID column, keeping track of where each row originates
            merged = df.merge(completed_ids, on='lat_lon_id', how='outer', indicator=True)

            # Filter out rows that come from the 'completed_ids' DataFrame
            df = merged[merged['_merge'] == 'left_only'].drop(columns='_merge')

        def get_street_view_info(longitude, latitude, proxies):
            results = panoids(latitude, longitude, proxies, )
            return results

        def worker(row):
            input_longitude = row.longitude
            input_latitude = row.latitude
            lat_lon_id = row.lat_lon_id
            id_dict = {column: getattr(row, column) for column in id_columns} if id_columns else {}
            return lat_lon_id, (input_longitude, input_latitude), get_street_view_info(input_longitude, input_latitude, self.proxies), id_dict

        # set lat_lon_id if it doesn't exist
        if 'lat_lon_id' not in df.columns:
            df['lat_lon_id'] = np.arange(1, len(df) + 1)
        results = []
        batch_size = 1000  # Modify this to a suitable value
        num_batches = (len(df) + batch_size - 1) // batch_size
        failed_rows = []
        
        # if there's no rows to process, return completed_ids
        if len(df) == 0:
            return completed_ids
        
        # if not, process the rows
        for i in tqdm(range(num_batches), desc=f"Getting pids by batch size {min(batch_size, len(df))}"):
            with ThreadPoolExecutor() as executor:
                batch_futures = {executor.submit(worker, row): row for row in df.iloc[i*batch_size : (i+1)*batch_size].itertuples()}
                for future in tqdm(as_completed(batch_futures), total=len(batch_futures), desc=f"Getting pids for batch #{i+1}"):
                    try:
                        lat_lon_id, (input_longitude, input_latitude), row_results, id_dict = future.result()
                        for result in row_results:
                            result['input_latitude'] = input_latitude
                            result['input_longitude'] = input_longitude
                            result['lat_lon_id'] = lat_lon_id
                            result.update(id_dict)
                            results.append(result)
                    except Exception as e:
                        print(f"Error: {e}")
                        failed_rows.append(batch_futures[future])  # Store the failed row

                # Save checkpoint for each batch
                if len(results) > 0:
                    pd.DataFrame(results).to_csv(f'{dir_cache_pids}/checkpoint_batch_{checkpoint_start_index+i+1}.csv', index=False)
                results = []  # Clear the results list for the next batch

        # Merge all checkpoints into a single dataframe
        results_df = pd.concat([pd.read_csv(checkpoint) for checkpoint in glob.glob(str(dir_cache_pids / '*.csv'))], ignore_index=True)

        # Retry failed rows
        if failed_rows:
            print("Retrying failed rows...")
            with ThreadPoolExecutor() as executor:
                retry_futures = {executor.submit(worker, row): row for row in failed_rows}
                for future in tqdm(as_completed(retry_futures), total=len(retry_futures), desc="Retrying failed rows"):
                    try:
                        lat_lon_id, (input_longitude, input_latitude), row_results, id_dict = future.result()
                        for result in row_results:
                            result['input_latitude'] = input_latitude
                            result['input_longitude'] = input_longitude
                            result['lat_lon_id'] = lat_lon_id
                            result.update(id_dict)
                            results.append(result)
                    except Exception as e:
                        print(f"Failed again: {e}")

            # Save the results of retried rows as another checkpoint
            if len(results) > 0:
                pd.DataFrame(results).to_csv(f'{dir_cache_pids}/checkpoint_retry.csv', index=False)
                # Merge the retry checkpoint into the final dataframe
                retry_df = pd.read_csv(f'{dir_cache_pids}/checkpoint_retry.csv')
                results_df = pd.concat([results_df, retry_df], ignore_index=True)

        # now save results_df as a new cache after dropping lat_lon_id
        results_df = results_df.drop(columns='lat_lon_id')
        # drop duplicates in panoid and id_columns
        results_df = results_df.drop_duplicates(subset=['panoid'] + id_columns)
        results_df.to_csv(self.cache_pids_raw, index=False)

        # delete the cache directory
        if dir_cache_pids.exists():
            shutil.rmtree(dir_cache_pids)
        return results_df

    def _get_pids_from_gdf(self, gdf, **kwargs):  
        if self.cache_lat_lon.exists():
            df = pd.read_csv(self.cache_lat_lon)
            print("The lat and lon have been read from the cache")
        else:
            if gdf.crs is None:
                gdf = gdf.set_crs('EPSG:4326')
            elif gdf.crs != 'EPSG:4326':
                # convert to EPSG:4326
                gdf = gdf.to_crs('EPSG:4326')
            # read shapefile
            gp = GeoProcessor(gdf, distance=self.distance, grid=self.grid, grid_size=self.grid_size, **kwargs)
            df = gp.get_lat_lon()
            df['lat_lon_id'] = np.arange(1, len(df) + 1)
            # save df to cache
            df.to_csv(self.cache_lat_lon, index=False)

        if self.cache_pids_raw.exists():
            print("The raw panorama IDs have been read from the cache")
            results_df = pd.read_csv(self.cache_pids_raw)
        else:
            # Use _get_pids_from_df to get pids from df
            results_df = self._get_pids_from_df(df, kwargs["id_columns"])

        # Check if lat and lon are within input polygons
        polygons = gpd.GeoSeries([geom for geom in gdf['geometry'] if geom.type in ['Polygon', 'MultiPolygon']])

        # the rest is only for polygons, so return results_df if there's no polygons
        if len(polygons) == 0:
            return results_df

        # Convert lat, lon to Points and create a GeoSeries
        points = gpd.GeoSeries([Point(lon, lat) for lon, lat in zip(results_df['lon'], results_df['lat'])])

        # Create a GeoDataFrame with the points and an index column
        points_gdf = gpd.GeoDataFrame(geometry=points, crs=gdf.crs)
        points_gdf['index'] = range(len(points_gdf))

        # Create a spatial index on the polygons GeoSeries
        polygons_sindex = polygons.sindex

        # Function to check whether a point is within any polygon
        def is_within_polygon(point):
            possible_matches_index = list(polygons_sindex.intersection(point.bounds))
            possible_matches = polygons.iloc[possible_matches_index]
            precise_matches = possible_matches.contains(point)
            return precise_matches.any()

        # Add progress bar for within_polygon calculation
        with tqdm(total=len(points), desc="Checking points within polygons") as pbar:
            within_polygon = []
            for point in points_gdf['geometry']:
                within_polygon.append(is_within_polygon(point))
                pbar.update()

        results_df['within_polygon'] = within_polygon

        # Return only those points within polygons
        results_within_polygons_df = results_df[results_df['within_polygon']]
        # Drop the 'within_polygon' column
        results_within_polygons_df = results_within_polygons_df.drop(columns='within_polygon')
        return results_within_polygons_df

    def _get_raw_pids(self, **kwargs):
        if self.cache_pids_raw.exists():
            pid = pd.read_csv(self.cache_pids_raw)
            print("The raw panorama IDs have been read from the cache")
            return pid

        if kwargs['lat'] is not None and kwargs['lon'] is not None:
            pid = panoids(kwargs['lat'], kwargs['lon'], self.proxies)
            pid = pd.DataFrame(pid)
            # add input_lat and input_lon
            pid['input_latitude'] = kwargs['lat']
            pid['input_longitude'] = kwargs['lon']
        elif kwargs['input_csv_file'] != "":
            df = pd.read_csv(kwargs['input_csv_file'])
            df = standardize_column_names(df)
            if kwargs['buffer'] > 0:
                gdf = gpd.GeoDataFrame(df, geometry=gpd.points_from_xy(df.longitude, df.latitude), crs='EPSG:4326')
                gdf = create_buffer_gdf(gdf, kwargs['buffer'])
                pid = self._get_pids_from_gdf(gdf, **kwargs)
            else:
                pid = self._get_pids_from_df(df, kwargs['id_columns'])
        elif kwargs['input_shp_file'] != "":
            gdf = gpd.read_file(kwargs['input_shp_file'])
            if kwargs['buffer'] > 0:
                gdf = create_buffer_gdf(gdf, kwargs['buffer'])
            pid = self._get_pids_from_gdf(gdf, **kwargs)
        elif kwargs['input_place_name'] != "":
            print("Geocoding the input place name")
            gdf = ox.geocoder.geocode_to_gdf(kwargs['input_place_name'])
            # raise error if the input_place_name is not found
            if len(gdf) == 0:
                raise ValueError("The input_place_name is not found. Please try another place name.")
            if kwargs['buffer'] > 0:
                gdf = create_buffer_gdf(gdf, kwargs['buffer'])
            pid = self._get_pids_from_gdf(gdf, **kwargs) 
        else:
            raise ValueError("Please input the lat and lon, csv file, or shapefile.")

        return pid

    def _get_pids(self, path_pid, **kwargs):
        id_columns = kwargs['id_columns']
        if id_columns is not None:
            if isinstance(id_columns, str):
                id_columns = [id_columns.lower()]
            elif isinstance(id_columns, list):
                id_columns = [column.lower() for column in id_columns]
        else:
            id_columns = []
        # update id_columns
        kwargs['id_columns'] = id_columns
        
        # get raw pid
        pid = self._get_raw_pids(**kwargs)
        
        if kwargs["augment_metadata"] & (self.gsv_api_key != None):
            pid = self._augment_metadata(pid)
        elif kwargs["augment_metadata"] & (self.gsv_api_key == None):
            raise ValueError("Please set the gsv api key by calling the gsv_api_key method.")
        pid.to_csv(path_pid, index=False)
        print("The panorama IDs have been saved to {}".format(path_pid)) 

    def _set_dirs(self, dir_output):
        # set dir_output as attribute and create the directory
        self.dir_output = Path(dir_output)
        self.dir_output.mkdir(parents=True, exist_ok=True)
        # set dir_cache as attribute and create the directory
        self.dir_cache = self.dir_output / "cache_zensvi"
        self.dir_cache.mkdir(parents=True, exist_ok=True)
        # set other cache directories
        self.cache_lat_lon = self.dir_cache / "lat_lon.csv"
        self.cache_pids_raw = self.dir_cache / "pids_raw.csv"
        self.cache_pids_augmented = self.dir_cache / "pids_augemented.csv"
        
    def _filter_pids_date(self, pid_df, start_date, end_date):
        # create a temporary column date from year and month
        pid_df['date'] = pid_df['year'].astype(str) + "-" + pid_df['month'].astype(str)
        # convert to datetime
        pid_df['date'] = pd.to_datetime(pid_df['date'], format="%Y-%m")
        # check if start_date and end_date are in the correct format with regex. If not, raise error
        if start_date is not None:
            try:
                start_date = datetime.datetime.strptime(start_date, "%Y-%m-%d")
            except ValueError:
                raise ValueError("Incorrect start_date format, should be YYYY-MM-DD")
        if end_date is not None:
            try:
                end_date = datetime.datetime.strptime(end_date, "%Y-%m-%d")
            except ValueError:
                raise ValueError("Incorrect end_date format, should be YYYY-MM-DD")
        # if start_date is not None, filter out the rows with date < start_date
        pid_df = pid_df[pid_df['date'] >= start_date] if start_date is not None else pid_df
        # if end_date is not None, filter out the rows with date > end_date
        pid_df = pid_df[pid_df['date'] <= end_date] if end_date is not None else pid_df
        # drop the temporary column date
        pid_df = pid_df.drop(columns='date')
        return pid_df
    
    def download_svi(self, dir_output: str, path_pid: str = None, zoom: int = 2, h_tiles: int = 4, v_tiles: int = 2, 
                      cropped: bool = False, full: bool = True, lat: float = None, lon: float = None, 
                      input_csv_file: str = "", input_shp_file: str = "", input_place_name: str = "", 
                      id_columns: Union[str, List[str]] = None, buffer: int = 0, augment_metadata: bool = False, 
                      batch_size: int = 1000, update_pids: bool = False, start_date = None, end_date = None, **kwargs) -> None:
        """
        Downloads street view images.

        Args:
            dir_output (str): The output directory.
            path_pid (str, optional): The path to the panorama ID file. Defaults to None.
            zoom (int, optional): The zoom level for the images. Defaults to 2.
            h_tiles (int, optional): The number of horizontal tiles. Defaults to 4.
            v_tiles (int, optional): The number of vertical tiles. Defaults to 2.
            cropped (bool, optional): Whether to crop the images. Defaults to False.
            full (bool, optional): Whether to download full images. Defaults to True.
            lat (float, optional): The latitude for the images. Defaults to None.
            lon (float, optional): The longitude for the images. Defaults to None.
            input_csv_file (str, optional): The input CSV file. Defaults to "".
            input_shp_file (str, optional): The input shapefile. Defaults to "".
            input_place_name (str, optional): The input place name. Defaults to "".
            id_columns (Union[str, List[str]], optional): The ID columns. Defaults to None.
            buffer (int, optional): The buffer size. Defaults to 0.
            augment_metadata (bool, optional): Whether to augment the metadata. Defaults to False.
            batch_size (int, optional): The batch size for downloading. Defaults to 1000.
            update_pids (bool, optional): Whether to update the panorama IDs. Defaults to False.
            start_date (str, optional): The start date for the panorama IDs. Format is isoformat (YYYY-MM-DD). Defaults to None.
            end_date (str, optional): The end date for the panorama IDs. Format is isoformat (YYYY-MM-DD). Defaults to None.
            **kwargs: Additional keyword arguments.

        Returns:
            None
        """
        # set necessary directories
        self._set_dirs(dir_output)
        
        # call _get_pids function first if path_pid is None
        if (path_pid is None) & (self.cache_pids_augmented.exists() == False):
            print("Getting pids...")
            path_pid = self.dir_output / "gsv_pids.csv"
            if path_pid.exists() & (update_pids == False):
                print("update_pids is set to False. So the following csv file will be used: {}".format(path_pid))
            else:
                self._get_pids(path_pid, lat=lat, lon=lon,
                            input_csv_file=input_csv_file, input_shp_file = input_shp_file, input_place_name = input_place_name, 
                            id_columns=id_columns, buffer = buffer, augment_metadata=augment_metadata, **kwargs)
        elif self.cache_pids_augmented.exists():
            # copy the cache pids_augmented to path_pid
            path_pid = self.dir_output / "gsv_pids.csv"
            shutil.copy2(self.cache_pids_augmented, path_pid)
            print("The augmented panorama IDs have been saved to {}".format(path_pid))
        # Horizontal Google Street View tiles
        # zoom 3: (8, 4); zoom 5: (26, 13) zoom 2: (4, 2) zoom 1: (2, 1);4:(8,16)
        # zoom = 2
        # h_tiles = 4  # 26
        # v_tiles = 2  # 13
        # cropped = False
        # full = True
        # create a folder within self.dir_output
        self.panorama_output = self.dir_output / "gsv_panorama"
        self.panorama_output.mkdir(parents=True, exist_ok=True)
        
        panoids = self._read_pids(path_pid, start_date, end_date)
        
        if len(panoids) == 0:
            print("There is no panorama ID to download")
            return
        else:
            panoids_rest = self._check_already(panoids)

        if len(panoids_rest) > 0:
            UAs = random.choices(self.user_agents, k = len(panoids_rest))
            ImageTool.dwl_multiple(panoids_rest, zoom, v_tiles, h_tiles, self.panorama_output, UAs, self.proxies, cropped, full, batch_size = batch_size, log_path=self.log_path)
        else:
            print("All images have been downloaded")
        
        # delete the cache directory
        if self.dir_cache.exists():
            shutil.rmtree(self.dir_cache)
            print("The cache directory has been deleted")


class MLYDownloader(BaseDownloader):
    def __init__(self, mly_api_key, log_path = None, distance = 1, grid = False, grid_size = 1, max_workers = None):
        super().__init__(log_path, distance, grid, grid_size) 
        self._mly_api_key = mly_api_key
        self._max_workers = max_workers
        mly.set_access_token(self.mly_api_key)
        
    @property
    def mly_api_key(self):
        return self._mly_api_key
    @mly_api_key.setter
    def mly_api_key(self,mly_api_key):
        self._mly_api_key = mly_api_key
    
    @property
    def max_workers(self):
        return self._max_workers
    @max_workers.setter
    def max_workers(self,max_workers):
        self._max_workers = max_workers
    
    def _read_pids(self, path_pid):
        pid_df = pd.read_csv(path_pid)
        # get unique pids (ie "id" columns) as a list
        pids = pid_df["id"].astype(int).unique().tolist()
        return pids

    def _set_dirs(self, dir_output):
        # set dir_output as attribute and create the directory
        self.dir_output = Path(dir_output)
        self.dir_output.mkdir(parents=True, exist_ok=True)
        self.pids_url = self.dir_output / "pids_urls.csv"
        # set dir_cache as attribute and create the directory
        self.dir_cache = self.dir_output / "cache_zensvi"
        self.dir_cache.mkdir(parents=True, exist_ok=True)
        # set other cache directories
        self.cache_lat_lon = self.dir_cache / "lat_lon.csv"
        self.cache_pids_raw = self.dir_cache / "pids_raw.csv"
    
    def _get_pids_from_df(self, df, id_columns, **kwargs):
        # 1. Create a new directory called "pids" to store each batch pids
        dir_cache_pids = self.dir_cache / 'raw_pids'
        dir_cache_pids.mkdir(parents=True, exist_ok=True)

        # 2. Load all the checkpoint csv files
        checkpoints = glob.glob(str(dir_cache_pids / '*.csv'))
        checkpoint_start_index = len(checkpoints)
        if checkpoint_start_index > 0:
            dataframes = []
            for checkpoint in checkpoints:
                try:
                    df_checkpoint = pd.read_csv(checkpoint)
                    dataframes.append(df_checkpoint)
                except pd.errors.EmptyDataError:
                    print(f"Warning: {checkpoint} is empty and has been skipped.")
                    continue
            completed_rows = pd.concat(dataframes, ignore_index=True)

            completed_ids = completed_rows['lat_lon_id'].drop_duplicates()

            # Merge on the ID column, keeping track of where each row originates
            merged = df.merge(completed_ids, on='lat_lon_id', how='outer', indicator=True)

            # Filter out rows that come from the 'completed_ids' DataFrame
            df = merged[merged['_merge'] == 'left_only'].drop(columns='_merge')

        def get_street_view_info(latitude, longitude, **kwargs):
            results = mly.get_image_close_to(latitude = latitude, longitude = longitude, **kwargs).to_dict()["features"]
            return results

        def worker(row, **kwargs):
            input_latitude = row.latitude
            input_longitude = row.longitude
            lat_lon_id = row.lat_lon_id
            id_dict = {column: getattr(row, column) for column in id_columns} if id_columns else {}
            return lat_lon_id, (input_latitude, input_longitude), get_street_view_info(input_latitude, input_longitude, **kwargs), id_dict

        # set lat_lon_id if it doesn't exist
        if 'lat_lon_id' not in df.columns:
            df['lat_lon_id'] = np.arange(1, len(df) + 1)
        results = []
        batch_size = 1000  # Modify this to a suitable value
        num_batches = (len(df) + batch_size - 1) // batch_size
        failed_rows = []
        
        # if there's no rows to process, return completed_ids
        if len(df) == 0:
            return completed_ids
        
        # if not, process the rows
        for i in tqdm(range(num_batches), desc=f"Getting pids by batch size {min(batch_size, len(df))}"):
            with ThreadPoolExecutor(max_workers=self._max_workers) as executor:
                batch_futures = {executor.submit(worker, row, **kwargs): row for row in df.iloc[i*batch_size : (i+1)*batch_size].itertuples()}
                for future in tqdm(as_completed(batch_futures), total=len(batch_futures), desc=f"Getting pids for batch #{i+1}"):
                    try:
                        lat_lon_id, (input_longitude, input_latitude), row_results, id_dict = future.result()
                        for result in row_results:
                            result['properties']['input_latitude'] = input_latitude
                            result['properties']['input_longitude'] = input_longitude
                            result['properties']['lat_lon_id'] = lat_lon_id
                            result['properties'].update(id_dict)
                            results.append(result)
                    except Exception as e:
                        print(f"Error: {e}")
                        failed_rows.append(batch_futures[future])  # Store the failed row

                # Save checkpoint for each batch
                if len(results) > 0:
                    # convert to geodataframe
                    results_gdf = gpd.GeoDataFrame.from_features(results)
                    results_gdf['lon'] = results_gdf['geometry'].apply(lambda geom: geom.x)
                    results_gdf['lat'] = results_gdf['geometry'].apply(lambda geom: geom.y)
                    # drop geometry column
                    results_gdf = results_gdf.drop(columns='geometry')  
                    pd.DataFrame(results_gdf).to_csv(f'{dir_cache_pids}/checkpoint_batch_{checkpoint_start_index+i+1}.csv', index=False)
                results = []  # Clear the results list for the next batch

        # Merge all checkpoints into a single dataframe
        checkpoints = glob.glob(str(dir_cache_pids / '*.csv'))
        if len(checkpoints) > 0:
            results_df = pd.concat([pd.read_csv(checkpoint) for checkpoint in checkpoints], ignore_index=True)

        # Retry failed rows
        if failed_rows:
            print("Retrying failed rows...")
            with ThreadPoolExecutor() as executor:
                retry_futures = {executor.submit(worker, row, **kwargs): row for row in failed_rows}
                for future in tqdm(as_completed(retry_futures), total=len(retry_futures), desc="Retrying failed rows"):
                    try:
                        lat_lon_id, (input_longitude, input_latitude), row_results, id_dict = future.result()
                        for result in row_results:
                            result['properties']['input_latitude'] = input_latitude
                            result['properties']['input_longitude'] = input_longitude
                            result['properties']['lat_lon_id'] = lat_lon_id
                            result['properties'].update(id_dict)
                            results.append(result)
                    except Exception as e:
                        print(f"Failed again: {e}")

            # Save the results of retried rows as another checkpoint
            if len(results) > 0:
                # convert to geodataframe
                results_gdf = gpd.GeoDataFrame.from_features(results)
                results_gdf['lon'] = results_gdf['geometry'].apply(lambda geom: geom.x)
                results_gdf['lat'] = results_gdf['geometry'].apply(lambda geom: geom.y)
                # drop geometry column
                results_gdf = results_gdf.drop(columns='geometry')   
                pd.DataFrame(results).to_csv(f'{dir_cache_pids}/checkpoint_retry.csv', index=False)
                # Merge the retry checkpoint into the final dataframe
                retry_df = pd.read_csv(f'{dir_cache_pids}/checkpoint_retry.csv')
                results_df = pd.concat([results_df, retry_df], ignore_index=True)
        
        if checkpoints == [] and failed_rows == []:
            print("There is no panorama ID to download")
            return

        # now save results_df as a new cache after dropping lat_lon_id and drop duplicates in panoid
        results_df = results_df.drop(columns='lat_lon_id')
        results_df = results_df.drop_duplicates(subset=['id'] + id_columns)
        results_df.to_csv(self.cache_pids_raw, index=False)

        # delete the cache directory
        if dir_cache_pids.exists():
            shutil.rmtree(dir_cache_pids)
        return results_df

    def _get_pids_from_gdf(self, gdf, mly_kwargs, **kwargs):
        if self.cache_lat_lon.exists():
            df = pd.read_csv(self.cache_lat_lon)
            print("The lat and lon have been read from the cache")
        else:
            if gdf.crs is None:
                gdf = gdf.set_crs('EPSG:4326')
            elif gdf.crs != 'EPSG:4326':
                # convert to EPSG:4326
                gdf = gdf.to_crs('EPSG:4326')
            # read shapefile
            gp = GeoProcessor(gdf, distance=self.distance, grid=self.grid, grid_size=self.grid_size, **kwargs)
            df = gp.get_lat_lon()
            df['lat_lon_id'] = np.arange(1, len(df) + 1)
            # save df to cache
            df.to_csv(self.cache_lat_lon, index=False)

        if self.cache_pids_raw.exists():
            print("The raw panorama IDs have been read from the cache")
            results_df = pd.read_csv(self.cache_pids_raw)
        else:
            # Use _get_pids_from_df to get pids from df
            results_df = self._get_pids_from_df(df, kwargs["id_columns"], **mly_kwargs)
            
        if not isinstance(results_df, pd.DataFrame):
            return
        
        # Check if lat and lon are within input polygons
        polygons = gpd.GeoSeries([geom for geom in gdf['geometry'] if geom.type in ['Polygon', 'MultiPolygon']])

        # the rest is only for polygons, so return results_df if there's no polygons
        if len(polygons) == 0:
            return results_df
        
        # Convert lat, lon to Points and create a GeoSeries
        points = gpd.GeoSeries([Point(lon, lat) for lon, lat in zip(results_df['lon'], results_df['lat'])])

        # Create a GeoDataFrame with the points and an index column
        points_gdf = gpd.GeoDataFrame(geometry=points, crs=gdf.crs)
        points_gdf['index'] = range(len(points_gdf))

        # Create a spatial index on the polygons GeoSeries
        polygons_sindex = polygons.sindex

        # Function to check whether a point is within any polygon
        def is_within_polygon(point):
            possible_matches_index = list(polygons_sindex.intersection(point.bounds))
            possible_matches = polygons.iloc[possible_matches_index]
            precise_matches = possible_matches.contains(point)
            return precise_matches.any()

        # Add progress bar for within_polygon calculation
        with tqdm(total=len(points), desc="Checking points within polygons") as pbar:
            within_polygon = []
            for point in points_gdf['geometry']:
                within_polygon.append(is_within_polygon(point))
                pbar.update()

        results_df['within_polygon'] = within_polygon

        # Return only those points within polygons
        results_within_polygons_df = results_df[results_df['within_polygon']]
        # Drop the 'within_polygon' column
        results_within_polygons_df = results_within_polygons_df.drop(columns='within_polygon')
        return results_within_polygons_df
        
    def _get_raw_pids(self, **kwargs):
        mly_allowed_keys = {'fields', 'zoom', 'radius', 'image_type', 'min_captured_at', 'max_captured_at', 'organization_id'} 
        mly_kwargs = {k: v for k, v in kwargs.items() if k in mly_allowed_keys}
        if self.cache_pids_raw.exists():
            pid = pd.read_csv(self.cache_pids_raw)
            print("The raw panorama IDs have been read from the cache")
            return pid

        if kwargs['lat'] is not None and kwargs['lon'] is not None:
            pid = mly.get_image_close_to(latitude = kwargs['lat'], longitude = kwargs['lon'], **mly_kwargs)
            # Convert to DataFrame
            pid = gpd.GeoDataFrame.from_features(pid.to_dict()['features'])
            pid['lon'] = pid['geometry'].apply(lambda geom: geom.x)
            pid['lat'] = pid['geometry'].apply(lambda geom: geom.y)
            # drop geometry column
            pid = pid.drop(columns='geometry')
            pid = pd.DataFrame(pid)
            # add input_lat and input_lon
            pid['input_latitude'] = kwargs['lat']
            pid['input_longitude'] = kwargs['lon']
        elif kwargs['input_csv_file'] != "":
            df = pd.read_csv(kwargs['input_csv_file'])
            df = standardize_column_names(df)
            if kwargs['buffer'] > 0:
                gdf = gpd.GeoDataFrame(df, geometry=gpd.points_from_xy(df.longitude, df.latitude), crs='EPSG:4326')
                gdf = create_buffer_gdf(gdf, kwargs['buffer'])
                pid = self._get_pids_from_gdf(gdf, mly_kwargs, **kwargs)
            else:
                pid = self._get_pids_from_df(df, kwargs['id_columns'], mly_kwargs)
        elif kwargs['input_shp_file'] != "":
            gdf = gpd.read_file(kwargs['input_shp_file'])
            if kwargs['buffer'] > 0:
                gdf = create_buffer_gdf(gdf, kwargs['buffer'])
            pid = self._get_pids_from_gdf(gdf, mly_kwargs, **kwargs)
        elif kwargs['input_place_name'] != "":
            print("Geocoding the input place name")
            gdf = ox.geocoder.geocode_to_gdf(kwargs['input_place_name'])
            # raise error if the input_place_name is not found
            if len(gdf) == 0:
                raise ValueError("The input_place_name is not found. Please try another place name.")
            if kwargs['buffer'] > 0:
                gdf = create_buffer_gdf(gdf, kwargs['buffer'])
            pid = self._get_pids_from_gdf(gdf, mly_kwargs, **kwargs) 
        else:
            raise ValueError("Please input the lat and lon, csv file, or shapefile.")

        return pid
    
    def _filter_pids_date(self, pid_df, start_date, end_date):
        # create a temporary column date from captured_at (milliseconds from Unix epoch)
        pid_df['date'] = pd.to_datetime(pid_df['captured_at'], unit='ms')
        # check if start_date and end_date are in the correct format with regex. If not, raise error
        if start_date is not None:
            try:
                start_date = datetime.datetime.strptime(start_date, "%Y-%m-%d")
            except ValueError:
                raise ValueError("Incorrect start_date format, should be YYYY-MM-DD")
        if end_date is not None:
            try:
                end_date = datetime.datetime.strptime(end_date, "%Y-%m-%d")
            except ValueError:
                raise ValueError("Incorrect end_date format, should be YYYY-MM-DD")
        # if start_date is not None, filter out the rows with date < start_date
        pid_df = pid_df[pid_df['date'] >= start_date] if start_date is not None else pid_df
        # if end_date is not None, filter out the rows with date > end_date
        pid_df = pid_df[pid_df['date'] <= end_date] if end_date is not None else pid_df
        # drop the temporary column date
        pid_df = pid_df.drop(columns='date')
        return pid_df

    def _get_pids(self, path_pid, **kwargs):
        id_columns = kwargs['id_columns']
        if id_columns is not None:
            if isinstance(id_columns, str):
                id_columns = [id_columns.lower()]
            elif isinstance(id_columns, list):
                id_columns = [column.lower() for column in id_columns]
        else:
            id_columns = []
        # update id_columns
        kwargs['id_columns'] = id_columns
        
        # get raw pid
        pid = self._get_raw_pids(**kwargs)
        
        if pid is None:
            print("There is no panorama ID to download")
            return

        # # Assuming that df is your DataFrame and that 'geometry' column contains Point objects
        # # convert the geometry column into shapely geometry objects
        # pid['geometry'] = pid['geometry'].apply(wkt.loads)
        # pid['lon'] = pid['geometry'].apply(lambda geom: geom.x)
        # pid['lat'] = pid['geometry'].apply(lambda geom: geom.y)
        # # drop geometry column
        # pid = pid.drop(columns='geometry')
        # move the "id" column to the first column
        pid = pid[["id"] + [col for col in pid.columns if col != "id"]]

        pid.to_csv(path_pid, index=False)
        print("The panorama IDs have been saved to {}".format(path_pid)) 
    
    def _get_urls_mly(self, path_pid, resolution=1024):
        # check if seld.cache_pids_urls exists
        if self.pids_url.exists():
            print("The panorama URLs have been read from the cache")
            return
        
        dir_cache_urls = self.dir_cache / 'urls'
        dir_cache_urls.mkdir(parents=True, exist_ok=True)

        checkpoints = glob.glob(str(dir_cache_urls / '*.csv'))
        checkpoint_start_index = len(checkpoints)

        panoids = set(self._read_pids(path_pid))  # Convert to set for faster operations
        if len(panoids) == 0:
            print("There is no panorama ID to download")
            return

        # Read all panoids from the checkpoint files
        completed_panoids = set()  # Use set for faster operations
        for checkpoint in checkpoints:
            try:
                df_checkpoint = pd.read_csv(checkpoint)
                completed_panoids.update(df_checkpoint['id'].tolist())
            except pd.errors.EmptyDataError:
                print(f"Warning: {checkpoint} is empty and has been skipped.")
                continue

        # Filter out the panoids that have already been processed
        panoids = list(panoids - completed_panoids)  # Subtract sets and convert back to list
        if len(panoids) == 0:
            print("All images have been downloaded")
            return

        def worker(panoid, resolution):
            url = mly.image_thumbnail(panoid, resolution=resolution)
            return panoid, url

        results = {}
        batch_size = 1000  # Modify this to a suitable value
        num_batches = (len(panoids) + batch_size - 1) // batch_size

        for i in tqdm(range(num_batches), desc=f"Getting urls by batch size {min(batch_size, len(panoids))}"):
            with ThreadPoolExecutor() as executor:
                batch_futures = {executor.submit(worker, panoid, resolution): panoid for panoid in panoids[i*batch_size : (i+1)*batch_size]}
                
                for future in tqdm(as_completed(batch_futures), total=len(batch_futures), desc=f"Getting urls for batch #{i+1}"):
                    current_panoid = batch_futures[future]
                    try:
                        panoid, url = future.result()
                        results[panoid] = url
                    except Exception as e:
                        print(f"Error: {e}")
                        self._log_write(current_panoid)
                        continue

            if len(results) > 0:
                pd.DataFrame.from_dict(results, orient='index').reset_index().rename(columns={'index': 'id', 0: 'url'}).to_csv(f'{dir_cache_urls}/checkpoint_batch_{checkpoint_start_index+i+1}.csv', index=False)
            results = {}

        # Merge all checkpoints into a single dataframe
        results_df = pd.concat([pd.read_csv(checkpoint) for checkpoint in glob.glob(str(dir_cache_urls / '*.csv'))], ignore_index=True)
        results_df.to_csv(self.pids_url, index=False)

        if dir_cache_urls.exists():
            shutil.rmtree(dir_cache_urls)


    def _download_images_mly(self, path_pid, cropped, batch_size, start_date, end_date):
        checkpoints = glob.glob(str(self.panorama_output / '**/*.png'), recursive=True)

        # Read already downloaded images and convert to ids
        downloaded_ids = set([Path(file_path).stem for file_path in checkpoints])  # Use set for faster operations

        pid_df = pd.read_csv(path_pid)
        pid_df["id"] = pid_df["id"].astype(int)
        urls_df = pd.read_csv(self.pids_url)
        urls_df["id"] = urls_df["id"].astype(int)
        # merge pid_df and urls_df
        urls_df = urls_df.merge(pid_df, on='id', how='left')
        # filter out the rows by date
        urls_df = self._filter_pids_date(urls_df, start_date, end_date)

        # Filter out the ids that have already been processed
        urls_df = urls_df[~urls_df['id'].isin(downloaded_ids)]  # Use isin for efficient operation

        def worker(row, output_dir, cropped):
            url, panoid = row.url, row.id
            user_agent = random.choice(self.user_agents)
            proxy = random.choice(self.proxies)

            image_name = f'{panoid}.png'  # Use id for file name
            image_path = output_dir / image_name
            try:
                response = requests.get(url, headers=user_agent, proxies=proxy, timeout=10)
                if response.status_code == 200:
                    with open(image_path, 'wb') as f:
                        f.write(response.content)
                    
                    if cropped:
                        img = Image.open(image_path)
                        w, h = img.size
                        img_cropped = img.crop((0, 0, w, h // 2))
                        img_cropped.save(image_path)

                else:
                    self._log_write(panoid)
            except Exception as e:
                self._log_write(panoid)
                print(f"Error: {e}" )

        num_batches = (len(urls_df) + batch_size - 1) // batch_size

        # Calculate current highest batch number
        existing_batches = glob.glob(str(self.panorama_output / "batch_*"))
        existing_batch_numbers = [int(Path(batch).name.split('_')[-1]) for batch in existing_batches]
        start_batch_number = max(existing_batch_numbers, default=0)

        for i in tqdm(range(start_batch_number, start_batch_number + num_batches), desc=f"Downloading images by batch size {min(batch_size, len(urls_df))}"):
            # Create a new sub-folder for each batch
            batch_out_path = self.panorama_output / f"batch_{i+1}"
            batch_out_path.mkdir(exist_ok=True)
            
            with ThreadPoolExecutor() as executor:
                batch_futures = {executor.submit(worker, row, batch_out_path, cropped): row.id for row in urls_df.iloc[i*batch_size : (i+1)*batch_size].itertuples()}
                for future in tqdm(as_completed(batch_futures), total=len(batch_futures), desc=f"Downloading images for batch #{i+1}"):
                    try:
                        future.result()
                    except Exception as e:
                        print(f"Error: {e}")

                
    def download_svi(self, dir_output, path_pid = None, lat=None, lon=None, input_csv_file="", input_shp_file = "", input_place_name =
                    "", id_columns=None, buffer = 0, update_pids = False, resolution = 1024, cropped = False, batch_size = 1000,
                    start_date = None, end_date = None, **kwargs):
        # set necessary directories
        self._set_dirs(dir_output)
        
        # call _get_pids function first if path_pid is None
        if path_pid is None:
            print("Getting pids...")
            path_pid = self.dir_output / "mly_pids.csv"
            if path_pid.exists() & (update_pids == False):
                print("update_pids is set to False. So the following csv file will be used: {}".format(path_pid))
            else:
                self._get_pids(path_pid, lat=lat, lon=lon,
                            input_csv_file=input_csv_file, input_shp_file = input_shp_file, input_place_name = input_place_name, 
                            id_columns=id_columns, buffer = buffer, start_date = start_date, end_date = end_date, **kwargs)
        else:
            # check if the path_pid exists
            if path_pid.exists():
                print("The following csv file will be used: {}".format(path_pid))
            else:
                self._get_pids(path_pid, lat=lat, lon=lon,
                            input_csv_file=input_csv_file, input_shp_file = input_shp_file, input_place_name = input_place_name, 
                            id_columns=id_columns, buffer = buffer, **kwargs)
    
        # create a folder within self.dir_output
        self.panorama_output = self.dir_output / "mly_svi"
        self.panorama_output.mkdir(parents=True, exist_ok=True)
        
        # get urls
        if path_pid.exists():
            self._get_urls_mly(path_pid, resolution=resolution) 
            # download images
            self._download_images_mly(path_pid, cropped, batch_size, start_date, end_date)
        else: 
            print("There is no panorama ID to download within the given input parameters")
        
        # delete the cache directory
        if self.dir_cache.exists():
            shutil.rmtree(self.dir_cache)
            print("The cache directory has been deleted") 



# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.config.api.general
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module contains the class implementation of the
general metadata functionalities for the API v4 of Mapillary.

For more information, please check out https://www.mapillary.com/developer/api-documentation/.

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""


class General:
    """
    A general list of metadata API endpoints for API v4
    """

    @staticmethod
    def get_tile_metadata():
        """Root endpoint for metadata"""

        return "https://graph.mapillary.com/"

    @staticmethod
    def get_vector_tiles():
        """Root endpoint for vector tiles"""

        return "https://tiles.mapillary.com/"

    @staticmethod
    def get_image_type_tiles(x: float, y: float, z: float) -> str:
        """image_type tiles"""

        return f"https://tiles.mapillary.com/maps/vtp/mly1_public/2/{z}/{x}/{y}/"

    @staticmethod
    def get_computed_image_type_tiles(
        x: float,
        y: float,
        z: float,
    ) -> str:
        """Computed image_type tiles"""

        return (
            "https://tiles.mapillary.com/maps/vtp/mly1_computed_public/2"
            f"/{z}/{x}/{y}/"
        )

    @staticmethod
    def get_map_features_points_tiles(
        x: float,
        y: float,
        z: float,
    ) -> str:
        """Map features (points) tiles"""

        return (
            f"https://tiles.mapillary.com/maps/vtp/mly_map_feature_point/2"
            f"/{z}/{x}/{y}/"
        )

    @staticmethod
    def get_map_features_traffic_signs_tiles(
        x: float,
        y: float,
        z: float,
    ) -> str:
        """Map features (traffic signs) tiles"""

        return (
            "https://tiles.mapillary.com/maps/vtp/"
            f"mly_map_feature_traffic_sign/2/{z}/{x}/{y}"
        )



# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.config.__init__
~~~~~~~~~~~~~~~~~~~~~~~~~

This module contains the API v4 configuration and URL arguments.

Created to help in ease of use of the API v4.

# SOME NOTES FROM THE API

Over Authentication
!!!!!!!!!!!!!!!!!!!

1. All requests against https://graph.mapillary.com must be authorized. They require a client or
    user access tokens. Tokens can be sent in two ways,

    1. Using ?access_token=XXX query parameters. This is a preferred method for interacting with an
        vector tiles. Using this method is STRONGLY discouraged for sending user access tokens
    2. Using a header such as Authorization: OAuth XXX, where XXX is the token obtained either
        through the OAuth flow that your application implements or a client token from
        https://mapillary.com/dashboard/developers

For more information, please check out https://www.mapillary.com/developer/api-documentation/.

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""

from . import api  # noqa: F401
